import io

import pandas as pd
import paramiko
import pymysql
from sshtunnel import SSHTunnelForwarder

from viadot.config import local_config
from viadot.exceptions import CredentialError
from viadot.sources.base import Source


class MySQL(Source):
    def __init__(
        self,
        config_key: str = None,
        credentials: dict = None,
        *args,
        **kwargs,
    ):
        """
        A class for interacting with DuckDB.

        Args:
            config_key (str, optional): The key inside local config containing the config.
            User can choose to use this or pass credentials directly to the `credentials`
            parameter. Defaults to None.
            credentials (dict, optional): Credentials for the connection. Defaults to None.
        """

        if config_key:
            config_credentials = local_config.get(config_key)

        credentials = credentials if credentials else config_credentials
        if credentials is None:
            raise CredentialError("Credentials not found.")

        super().__init__(*args, credentials=credentials, **kwargs)

    @property
    def con(self) -> pymysql.connect:
        """
        Return a new connection to the MySQL database.

        Returns:
            pymysql.connect: database connection.
        """

        if self.credentials.get("host") is None:
            host = "127.0.0.1"

        conn = pymysql.connect(
            host=host,
            user=self.credentials.get("sql_username"),
            passwd=self.credentials.get("sql_password"),
            db=self.credentials.get("sql_main_database"),
            port=self.credentials.get("sql_port"),
        )
        return conn

    def to_df(self, query: str) -> pd.DataFrame:
        """
        Get DataFrame from MySQL database by pandas and SQL query.

        Returns:
            pd.DataFrame: Pandas dataframe.
        """
        data = pd.read_sql_query(query, self.con)
        self.con.close()
        return data

    def connect_sql_ssh(
        self,
        query: str,
    ) -> pd.DataFrame:
        """
        Establish connection with database using Secure Shell Protocol (SSH) and
        get pandas DataFrame by proper SQL query.

        Returns:
            pd.DataFrame: Pandas dataframe.
        """

        if self.credentials.get("host") is None:
            host = "127.0.0.1"

        sql_hostname = self.credentials.get("sql_hostname")
        sql_username = self.credentials.get("sql_username")
        sql_password = self.credentials.get("sql_password")
        sql_db_name = self.credentials.get("sql_db_name")
        sql_port = self.credentials.get("sql_port")

        ssh_host = self.credentials.get("ssh_host")
        ssh_user = self.credentials.get("ssh_user")
        ssh_port = self.credentials.get("ssh_port")
        rsakey = self.credentials.get("rsakey")

        rsakey = self.credentials["rsakey"]
        ssh_pkey = paramiko.RSAKey.from_private_key(io.StringIO(rsakey))

        with SSHTunnelForwarder(
            (ssh_host, ssh_port),
            ssh_username=ssh_user,
            ssh_pkey=ssh_pkey,
            remote_bind_address=(sql_hostname, sql_port),
        ) as ssh_tunnel:
            conn = pymysql.connect(
                host=host,
                user=sql_username,
                passwd=sql_password,
                db=sql_db_name,
                port=ssh_tunnel.local_bind_port,
            )

            df = pd.read_sql_query(query, conn)
            conn.close()
            return df



import copy
import json
import os
from typing import List

import pandas as pd
from prefect import Task
from prefect.tasks.secrets import PrefectSecret
from prefect.utilities import logging
from prefect.utilities.tasks import defaults_from_attrs

from ..exceptions import ValidationError
from ..sources import Sharepoint
from .azure_key_vault import AzureKeyVaultSecret
from ..utils import add_viadot_metadata_columns

logger = logging.get_logger()


class SharepointToDF(Task):
    """
    Task for converting data from Sharepoint excel file to a pandas DataFrame.

    Args:
        path_to_file (str): Path to Excel file.
        url_to_file (str):  Link to a file on Sharepoint.
                        (e.g : https://{tenant_name}.sharepoint.com/sites/{folder}/Shared%20Documents/Dashboard/file). Defaults to None.
        nrows (int, optional): Number of rows to read at a time. Defaults to 50000.
        sheet_number (int): Sheet number to be extracted from file. Counting from 0, if None all sheets are axtracted. Defaults to None.
        validate_excel_file (bool, optional): Check if columns in separate sheets are the same. Defaults to False.
        if_empty (str, optional): What to do if query returns no data. Defaults to "warn".
        timeout(int, optional): The amount of time (in seconds) to wait while running this task before
            a timeout occurs. Defaults to 3600.

    Returns:
        pd.DataFrame: Pandas data frame
    """

    def __init__(
        self,
        path_to_file: str = None,
        url_to_file: str = None,
        nrows: int = 50000,
        sheet_number: int = None,
        validate_excel_file: bool = False,
        if_empty: str = "warn",
        timeout: int = 3600,
        *args,
        **kwargs,
    ):
        self.if_empty = if_empty
        self.path_to_file = path_to_file
        self.url_to_file = url_to_file
        self.nrows = nrows
        self.sheet_number = sheet_number
        self.validate_excel_file = validate_excel_file

        super().__init__(
            name="sharepoint_to_df",
            timeout=timeout,
            *args,
            **kwargs,
        )

    def __call__(self):
        """Download Sharepoint data to a DF"""
        super().__call__(self)

    def check_column_names(
        self, df_header: List[str] = None, header_to_compare: List[str] = None
    ) -> List[str]:
        """
        Check if column names in sheets are the same.

        Args:
            df_header (List[str]): Header of df from excel sheet.
            header_to_compare (List[str]): Header of df from previous excel sheet.

        Returns:
            list: list of columns
        """
        df_header_list = df_header.columns.tolist()
        if header_to_compare is not None:
            if df_header_list != header_to_compare:
                raise ValidationError("Columns in sheets are different")

        return df_header_list

    def df_replace_special_chars(self, df: pd.DataFrame):
        """
        Replace "\n" and "\t" with "".

        Args:
            df (pd.DataFrame): Pandas data frame to replace characters.

        Returns:
            df (pd.DataFrame): Pandas data frame

        """
        return df.replace(r"\n|\t", "", regex=True)

    def split_sheet(
        self,
        sheetname: str = None,
        nrows: int = None,
        chunks: List[pd.DataFrame] = None,
        **kwargs,
    ) -> List[pd.DataFrame]:
        """
        Split sheet by chunks.

        Args:
            sheetname (str): The sheet on which we iterate.
            nrows (int): Number of rows to read at a time.
            chunks(List[pd.DataFrame]): List of data in chunks.

        Returns:
            List[pd.DataFrame]: List of data frames
        """
        skiprows = 1
        logger.info(f"Worksheet: {sheetname}")
        temp_chunks = copy.deepcopy(chunks)
        i_chunk = 0
        while True:
            df_chunk = pd.read_excel(
                self.path_to_file,
                sheet_name=sheetname,
                nrows=nrows,
                skiprows=skiprows,
                header=None,
                **kwargs,
            )
            skiprows += nrows
            # When there is no data, we know we can break out of the loop.
            if df_chunk.empty:
                break
            else:
                logger.debug(f" - chunk {i_chunk+1} ({df_chunk.shape[0]} rows)")
                df_chunk["sheet_name"] = sheetname
                temp_chunks.append(df_chunk)
            i_chunk += 1
        return temp_chunks

    @defaults_from_attrs(
        "path_to_file",
        "url_to_file",
        "nrows",
        "sheet_number",
        "validate_excel_file",
    )
    @add_viadot_metadata_columns(source_name="Sharepoint")
    def run(
        self,
        path_to_file: str = None,
        url_to_file: str = None,
        nrows: int = 50000,
        validate_excel_file: bool = False,
        sheet_number: int = None,
        credentials_secret: str = None,
        vault_name: str = None,
        **kwargs,
    ) -> None:
        """
        Run Task ExcelToDF.

        Args:
            path_to_file (str): Path to Excel file. Defaults to None.
            url_to_file (str): Link to a file on Sharepoint. Defaults to None.
            nrows (int, optional): Number of rows to read at a time. Defaults to 50000.
            sheet_number (int): Sheet number to be extracted from file. Counting from 0, if None all sheets are axtracted. Defaults to None.
            validate_excel_file (bool, optional): Check if columns in separate sheets are the same. Defaults to False.
            credentials_secret (str, optional): The name of the Azure Key Vault secret containing a dictionary with
            ACCOUNT_NAME and Service Principal credentials (TENANT_ID, CLIENT_ID, CLIENT_SECRET). Defaults to None.
            vault_name (str, optional): The name of the vault from which to obtain the secret. Defaults to None.

        Returns:
            pd.DataFrame: Pandas data frame
        """
        if not credentials_secret:
            # attempt to read a default for the service principal secret name
            try:
                credentials_secret = PrefectSecret("SHAREPOINT_KV").run()
            except ValueError:
                pass

        if credentials_secret:
            credentials_str = AzureKeyVaultSecret(
                credentials_secret, vault_name=vault_name
            ).run()
            credentials = json.loads(credentials_str)

        self.path_to_file = path_to_file
        self.url_to_file = url_to_file
        path_to_file = os.path.basename(self.path_to_file)
        self.sheet_number = sheet_number

        s = Sharepoint(download_from_path=self.url_to_file, credentials=credentials)
        s.download_file(download_to_path=path_to_file)

        self.nrows = nrows
        excel = pd.ExcelFile(self.path_to_file)

        if self.sheet_number is not None:
            sheet_names_list = [excel.sheet_names[self.sheet_number]]
        else:
            sheet_names_list = excel.sheet_names

        header_to_compare = None
        chunks = []

        for sheetname in sheet_names_list:
            df_header = pd.read_excel(self.path_to_file, sheet_name=sheetname, nrows=0)

            if validate_excel_file:
                header_to_compare = self.check_column_names(
                    df_header, header_to_compare
                )

            chunks = self.split_sheet(sheetname, self.nrows, chunks)
            df_chunks = pd.concat(chunks)

            # Rename the columns to concatenate the chunks with the header.
            columns = {i: col for i, col in enumerate(df_header.columns.tolist())}
            last_column = len(columns)
            columns[last_column] = "sheet_name"

            df_chunks.rename(columns=columns, inplace=True)
            df = pd.concat([df_header, df_chunks])

        df = self.df_replace_special_chars(df)
        self.logger.info(f"Successfully converted data to a DataFrame.")
        return df



from viadot.config import local_config
from viadot.sources import Supermetrics


def test_connection():
    credentials = local_config.get("SUPERMETRICS")
    s = Supermetrics()
    google_ads_params = {
        "ds_id": "AW",
        "ds_accounts": ["1007802423"],
        "ds_user": credentials["USER"],
        "date_range_type": "last_month",
        "fields": [
            "Date",
            "Campaignname",
            "Clicks",
        ],
        "max_rows": 1,
    }
    df = s.query(google_ads_params).to_df()
    assert df.count()[0] > 0



import logging
import os
from typing import Any, Dict, List, Literal

import pandas as pd
from prefect import Flow
from prefect.run_configs import DockerRun
from prefect.tasks.secrets import PrefectSecret

from viadot.flows.aselite_to_adls import ASELiteToADLS
from viadot.task_utils import df_converts_bytes_to_int, df_to_csv
from viadot.tasks import AzureDataLakeUpload
from viadot.tasks.aselite import ASELiteToDF

TMP_FILE_NAME = "test_flow.csv"
MAIN_DF = None

df_task = ASELiteToDF()
file_to_adls_task = AzureDataLakeUpload()


def test_aselite_to_adls():
    credentials_secret = PrefectSecret("aselite").run()
    vault_name = PrefectSecret("AZURE_DEFAULT_KEYVAULT").run()

    query_designer = """SELECT TOP 10 [ID]
        ,[SpracheText]
        ,[SpracheKat]
        ,[SpracheMM]
        ,[KatSprache]
        ,[KatBasisSprache]
        ,[CodePage]
        ,[Font]
        ,[Neu]
        ,[Upd]
        ,[UpdL]
        ,[LosKZ]
        ,[AstNr]
        ,[KomKz]
        ,[RKZ]
        ,[ParentLanguageNo]
        ,[UPD_FIELD]
    FROM [UCRMDEV].[dbo].[CRM_00]"""

    flow = ASELiteToADLS(
        "Test flow",
        query=query_designer,
        sqldb_credentials_secret=credentials_secret,
        vault_name=vault_name,
        file_path=TMP_FILE_NAME,
        to_path="raw/supermetrics/mp/result_df_flow_at_des_m.csv",
        run_config=None,
    )

    result = flow.run()
    assert result.is_successful()

    MAIN_DF = pd.read_csv(TMP_FILE_NAME, delimiter="\t")

    assert isinstance(MAIN_DF, pd.DataFrame) == True

    assert MAIN_DF.shape == (10, 17)

    os.remove(TMP_FILE_NAME)



import os
from typing import List
from unittest import mock

import numpy as np
import pandas as pd
import prefect
import pytest

from viadot.task_utils import (
    add_ingestion_metadata_task,
    adls_bulk_upload,
    anonymize_df,
    chunk_df,
    df_clean_column,
    df_converts_bytes_to_int,
    df_get_data_types_task,
    df_map_mixed_dtypes_for_parquet,
    df_to_csv,
    df_to_parquet,
    dtypes_to_json_task,
    union_dfs_task,
    write_to_json,
)


class MockAzureUploadClass:
    def run(
        from_path: str = "",
        to_path: str = "",
        sp_credentials_secret: str = "",
        overwrite: bool = False,
    ) -> None:
        pass


def count_dtypes(dtypes_dict: dict = None, dtypes_to_count: List[str] = None) -> int:
    dtypes_counter = 0
    for v in dtypes_dict.values():
        if v in dtypes_to_count:
            dtypes_counter += 1
    return dtypes_counter


def test_add_ingestion_metadata_task():
    df = pd.DataFrame({"col1": [1, 2], "col2": [3, 4]})
    result = add_ingestion_metadata_task.run(df)
    assert "_viadot_downloaded_at_utc" in result.columns


def test_add_ingestion_metadata_task_empty():
    df = pd.DataFrame()
    result = add_ingestion_metadata_task.run(df)
    assert result.empty


def test_add_ingestion_metadata_task_no_data():
    df = pd.DataFrame({"col1": []})
    result = add_ingestion_metadata_task.run(df)
    assert "_viadot_downloaded_at_utc" in result.columns


def test_map_dtypes_for_parquet():
    df = pd.DataFrame(
        {
            "a": {0: 55.7, 1: "Hello", 2: "Hello"},
            "b": {0: "Start", 1: "Hello", 2: "Hello"},
            "w": {0: 679, 1: "Hello", 2: "Hello"},
            "x": {0: 1, 1: 2, 2: 444},
            "y": {0: 1.5, 1: 11.97, 2: 56.999},
            "z": {0: "Start", 1: 1, 2: "2021-01-01"},
        }
    )
    dtyps_dict = df_get_data_types_task.run(df)
    sum_of_dtypes = count_dtypes(dtyps_dict, ["Object", "String"])

    df_map = df_map_mixed_dtypes_for_parquet.run(df, dtyps_dict)
    dtyps_dict_mapped = df_get_data_types_task.run(df_map)
    sum_of_mapped_dtypes = count_dtypes(dtyps_dict_mapped, ["String"])

    assert sum_of_dtypes == sum_of_mapped_dtypes


def test_df_converts_bytes_to_int():
    data = {
        "ID": {0: 1, 1: 2, 2: 100, 3: 101, 4: 102},
        "SpracheText": {
            0: "TE_CATALOG_BASE_LANG",
            1: "TE_Docu",
            2: "TE_German",
            3: "TE_English",
            4: "TE_French",
        },
        "RKZ": {
            0: b"\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\r\x01\x00\x00\x00\x00\x00\x00\x04\x00\x00q\x9f#NV\x8dG\x00",
            1: b"\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\r\x01\x00\x00\x00\x00\x00\x00\x04\x00\x00<\xa0#NV\x8dG\x00",
            2: b"\r\xa3\x86\x01\x00\x01\x00\x00\x00\x00\x04\r\x01\x00\x00\x00\x00\x00\x00\x04\x00\x003\x9f#NV\x8dG\x00",
            3: b"\r\xa3\x86\x01\x00\x01\x00\x00\x00\x00\x04\r\x01\x00\x00\x00\x00\x00\x00\x04\x00\x00R\x9f#NV\x8dG\x00",
            4: b"\r\xa3\x86\x01\x00\x01\x00\x00\x00\x00\x04\r\x01\x00\x00\x00\x00\x00\x00\x04\x00\x00\xee\x9f#NV\x8dG\x00",
        },
    }

    df = pd.DataFrame.from_dict(data)
    test_df = df_converts_bytes_to_int.run(df)
    lst = test_df["RKZ"][0]
    is_int = all(isinstance(x, (int, int)) for x in lst)
    assert is_int == True


def test_chunk_df():
    df = pd.DataFrame(
        {
            "AA": [1, 2, 3, 4, 5],
            "BB": [11, 22, 33, 4, 5],
            "CC": [4, 5, 6, 4, 5],
            "DD": [44, 55, 66, 1, 2],
        }
    )
    res = chunk_df.run(df=df, size=2)
    assert len(res) == 3


def test_df_get_data_types_task():
    df = pd.DataFrame(
        {
            "a": {0: "ann", 1: "test", 2: "Hello"},
            "b": {0: 9, 1: "2021-01-01", 2: "Hello"},
            "w": {0: 679, 1: "Hello", 2: "Hello"},
            "x": {0: -1, 1: 2, 2: 444},
            "y": {0: 1.5, 1: 11.97, 2: 56.999},
            "z": {0: "2022-01-01", 1: "2021-11-01", 2: "2021-01-01"},
        }
    )
    res = df_get_data_types_task.run(df)
    assert res == {
        "a": "String",
        "b": "Object",
        "w": "Object",
        "x": "Integer",
        "y": "Float",
        "z": "Date",
    }


def test_df_to_csv():
    df = pd.DataFrame(
        {
            "a": {0: "a", 1: "b", 2: "c"},
            "b": {0: "a", 1: "b", 2: "c"},
            "w": {0: "a", 1: "b", 2: "c"},
        }
    )

    df_to_csv.run(df, "test.csv")
    result = pd.read_csv("test.csv", sep="\t")
    assert df.equals(result)
    os.remove("test.csv")


def test_df_to_parquet():
    df = pd.DataFrame(
        {
            "a": {0: "a", 1: "b", 2: "c"},
            "b": {0: "a", 1: "b", 2: "c"},
            "w": {0: "a", 1: "b", 2: "c"},
        }
    )

    df_to_parquet.run(df, "test.parquet")
    result = pd.read_parquet("test.parquet")
    assert df.equals(result)
    os.remove("test.parquet")


def test_union_dfs_task():
    df1 = pd.DataFrame(
        {
            "a": {0: "a", 1: "b", 2: "c"},
            "b": {0: "a", 1: "b", 2: "c"},
            "w": {0: "a", 1: "b", 2: "c"},
        }
    )
    df2 = pd.DataFrame(
        {
            "a": {0: "d", 1: "e"},
            "b": {0: "d", 1: "e"},
        }
    )
    list_dfs = []
    list_dfs.append(df1)
    list_dfs.append(df2)
    res = union_dfs_task.run(list_dfs)
    assert isinstance(res, pd.DataFrame)
    assert len(res) == 5


def test_dtypes_to_json_task():
    dtypes = {"country": "VARCHAR(100)", "sales": "FLOAT(24)"}
    dtypes_to_json_task.run(dtypes_dict=dtypes, local_json_path="dtypes.json")
    assert os.path.exists("dtypes.json")
    os.remove("dtypes.json")


def test_write_to_json():
    dict = {"name": "John", 1: [2, 4, 3]}
    write_to_json.run(dict, "dict.json")
    assert os.path.exists("dict.json")
    os.remove("dict.json")


def test_df_clean_column_all():
    data = {
        "col_1": ["a", "b\\r", "\tc", "d \r\n a"],
        "col_2": ["a", "b\\r", "\tc", "d \r\n a"],
    }
    expected_output = {
        "col_1": {0: "a", 1: "b", 2: "c", 3: "d  a"},
        "col_2": {0: "a", 1: "b", 2: "c", 3: "d  a"},
    }
    df = pd.DataFrame.from_dict(data)
    output = df_clean_column.run(df).to_dict()
    assert expected_output == output


def test_df_clean_column_defined():
    data = {
        "col_1": ["a", "b", "c", "d  a"],
        "col_2": ["a\t\r", "b\\r", "\tc", "d \r\n a"],
    }
    expected_output = {
        "col_1": {0: "a", 1: "b", 2: "c", 3: "d  a"},
        "col_2": {0: "a", 1: "b", 2: "c", 3: "d  a"},
    }
    df = pd.DataFrame.from_dict(data)
    output = df_clean_column.run(df, ["col_2"]).to_dict()
    assert output == expected_output


@mock.patch("viadot.task_utils.AzureDataLakeUpload", return_value=MockAzureUploadClass)
@pytest.mark.bulk
def test_adls_bulk_upload(mock_upload):
    file_names = ["random_1.csv", "random_2.csv"]

    adls_bulk_upload.run(file_names=file_names, adls_file_path="any/at/random")
    mock_upload.assert_called_once()


def test_anonymize_df_all():
    data = pd.DataFrame(
        {
            "date_col": ["2005-01-01", "2010-01-02", "2023-01-03"],
            "email": [
                "john.malkovich@bing.com",
                "hannah.montana@disney.com",
                "ilovedogs33@dyvenia.com",
            ],
            "name": ["John", "Hannah", "Patryk"],
            "last_name": ["Malkovich", "Montana", "Dyvenian"],
            "active": [1, 0, 0],
        }
    )

    expected_output = {
        "date_col": {0: "2005-01-01", 1: "2010-01-02", 2: "2023-01-03"},
        "email": {0: "***", 1: "***", 2: "***"},
        "name": {0: "***", 1: "***", 2: "***"},
        "last_name": {0: "***", 1: "***", 2: "***"},
        "active": {0: 1, 1: 0, 2: 0},
    }
    output = anonymize_df.run(data, ["email", "name", "last_name"]).to_dict()
    assert output == expected_output


def test_anonymize_df_hash_defined_threshold_date():
    data = pd.DataFrame(
        {
            "date_col": ["2005-01-01", "2010-01-02", "2023-01-03"],
            "email": [
                "john.malkovich@bing.com",
                "hannah.montana@disney.com",
                "ilovedogs33@dyvenia.com",
            ],
            "name": ["John", "Hannah", "Patryk"],
            "last_name": ["Malkovich", "Montana", "Dyvenian"],
            "active": [1, 0, 0],
        }
    )

    expected_output = {
        "date_col": {0: "2005-01-01", 1: "2010-01-02", 2: "2023-01-03"},
        "email": {
            0: hash("john.malkovich@bing.com"),
            1: hash("hannah.montana@disney.com"),
            2: "ilovedogs33@dyvenia.com",
        },
        "name": {0: hash("John"), 1: hash("Hannah"), 2: "Patryk"},
        "last_name": {0: hash("Malkovich"), 1: hash("Montana"), 2: "Dyvenian"},
        "active": {0: 1, 1: 0, 2: 0},
    }

    output = anonymize_df.run(
        data,
        ["email", "name", "last_name"],
        method="hash",
        date_column="date_col",
        days=2 * 365,
    ).to_dict()
    assert output == expected_output


def test_anonymize_df_various_date_formats():
    data = pd.DataFrame(
        {
            "date_col1": ["2021-01-01", "2022-01-02", "2023-01-03"],
            "date_col2": [
                "2021-01-01 10:00:00",
                "2022-01-02 11:30:00",
                "2023-01-03 12:45:00",
            ],
            "date_col3": ["01/01/2021", "02/01/2022", "03/01/2023"],
            "date_col4": ["20210101", "20220102", "20230103"],
            "date_col5": [
                "2021-01-01T20:17:46.384Z",
                "2022-01-02T20:17:46.384Z",
                "2023-01-03T20:17:46.384Z",
            ],
            "date_col6": [
                "2021-01-01T20:17:46.384",
                "2022-01-02T20:17:46.384",
                "2023-01-03T20:17:46.384",
            ],
            "email": [
                "john.malkovich@bing.com",
                "hannah.montana@disney.com",
                "ilovedogs33@dyvenia.com",
            ],
            "active": [1, 0, 0],
        }
    )

    output1 = anonymize_df.run(
        data, ["email"], value=None, date_column="date_col1", days=2 * 365
    ).to_dict()
    output2 = anonymize_df.run(
        data, ["email"], value=None, date_column="date_col2", days=2 * 365
    ).to_dict()
    output3 = anonymize_df.run(
        data, ["email"], value=None, date_column="date_col3", days=2 * 365
    ).to_dict()
    output4 = anonymize_df.run(
        data, ["email"], value=None, date_column="date_col4", days=2 * 365
    ).to_dict()
    output5 = anonymize_df.run(
        data, ["email"], value=None, date_column="date_col5", days=2 * 365
    ).to_dict()
    output6 = anonymize_df.run(
        data, ["email"], value=None, date_column="date_col6", days=2 * 365
    ).to_dict()
    assert output1 == output2 == output3 == output4 == output5 == output6


def test_wrong_column():
    data = pd.DataFrame(
        {
            "date_col": ["2005-01-01", "2010-01-02", "2023-01-03"],
            "email": [
                "john.malkovich@bing.com",
                "hannah.montana@disney.com",
                "ilovedogs33@dyvenia.com",
            ],
            "name": ["John", "Hannah", "Patryk"],
            "last_name": ["Malkovich", "Montana", "Dyvenian"],
            "active": [1, 0, 0],
        }
    )
    with pytest.raises(ValueError, match="column names"):
        anonymize_df.run(data, ["first_name", "last_name", "email"])


def test_wrong_method():
    data = pd.DataFrame(
        {
            "date_col": ["2005-01-01", "2010-01-02", "2023-01-03"],
            "email": [
                "john.malkovich@bing.com",
                "hannah.montana@disney.com",
                "ilovedogs33@dyvenia.com",
            ],
            "name": ["John", "Hannah", "Patryk"],
            "last_name": ["Malkovich", "Montana", "Dyvenian"],
            "active": [1, 0, 0],
        }
    )
    with pytest.raises(ValueError, match="Method not found"):
        anonymize_df.run(data, ["name", "last_name", "email"], method="anonymize")



import logging

import pandas as pd
import pytest

from viadot.tasks import BigQueryToDF

logger = logging.getLogger(__name__)
DATASET_NAME = "manigeo"
TABLE_NAME = "space"
CREDENTIALS_KEY = "BIGQUERY-TESTS"


def test_bigquery_to_df_success():
    bigquery_to_df_task = BigQueryToDF(
        query=f"SELECT * FROM `manifest-geode-341308.{DATASET_NAME}.{TABLE_NAME}`",
        credentials_key=CREDENTIALS_KEY,
    )
    df = bigquery_to_df_task.run()
    expected_column = ["my_value", "_viadot_source"]

    assert isinstance(df, pd.DataFrame)
    assert expected_column == list(df.columns)


def test_bigquery_to_df_wrong_table_name(caplog):
    bigquery_to_df_task = BigQueryToDF()
    with caplog.at_level(logging.WARNING):
        df = bigquery_to_df_task.run(
            dataset_name=DATASET_NAME,
            table_name="wrong_table_name",
            date_column_name="date",
            credentials_key=CREDENTIALS_KEY,
        )
    assert f"Returning empty data frame." in caplog.text
    assert df.empty


def test_bigquery_to_df_wrong_column_name(caplog):
    bigquery_to_df_task = BigQueryToDF(
        dataset_name=DATASET_NAME,
        table_name=TABLE_NAME,
        date_column_name="wrong_column_name",
        credentials_key=CREDENTIALS_KEY,
    )
    with caplog.at_level(logging.WARNING):
        df = bigquery_to_df_task.run()
    assert f"'wrong_column_name' column is not recognized." in caplog.text
    assert isinstance(df, pd.DataFrame)


def test_bigquery_to_df_wrong_query(caplog):
    bigquery_to_df_task = BigQueryToDF(
        query="SELECT * FROM table_name",
        credentials_key=CREDENTIALS_KEY,
    )
    with caplog.at_level(logging.WARNING):
        df = bigquery_to_df_task.run()
    assert f"The query is invalid. Please enter a valid query." in caplog.text
    assert df.empty



from datetime import datetime
from typing import Any, Dict, Literal

import pandas as pd
from prefect.utilities import logging

from viadot.config import local_config
from viadot.exceptions import APIError, CredentialError
from viadot.sources.base import Source
from viadot.utils import handle_api_response

logger = logging.get_logger()


class CustomerGauge(Source):
    API_URL = "https://api.eu.customergauge.com/v7/rest/sync/"

    def __init__(
        self,
        endpoint: Literal["responses", "non-responses"] = None,
        url: str = None,
        credentials: Dict[str, Any] = None,
    ):
        """
        A class to connect and download data using Customer Gauge API.
        Below is the documentation for each of this API's gateways:
            Responses gateway https://support.customergauge.com/support/solutions/articles/5000875861-get-responses
            Non-Responses gateway https://support.customergauge.com/support/solutions/articles/5000877703-get-non-responses

        Args:
            endpoint (Literal["responses", "non-responses"]): Indicate which endpoint to connect. Defaults to None.
            url (str, optional): Endpoint URL. Defaults to None.
            credentials (Dict[str, Any], optional): Credentials to connect with API containing client_id, client_secret. Defaults to None.

        Raises:
            ValueError: If endpoint is not provided or incorect.
            CredentialError: If credentials are not provided in local_config or directly as a parameter
        """
        self.endpoint = endpoint
        if endpoint is not None:
            if endpoint in ["responses", "non-responses"]:
                self.url = f"{self.API_URL}{endpoint}"
            else:
                raise ValueError(
                    "Incorrect endpoint name. Choose: 'responses' or 'non-responses'"
                )
        elif url is not None:
            self.url = url
        else:
            raise ValueError(
                "Provide endpoint name. Choose: 'responses' or 'non-responses'. Otherwise, provide URL"
            )

        if credentials is not None:
            self.credentials = credentials
        else:
            self.credentials = local_config.get("CustomerGauge")
            if self.credentials is None:
                raise CredentialError("Credentials not provided.")

        super().__init__(credentials=self.credentials)

    def get_token(self) -> str:
        """
        Gets Bearer Token using POST request method.

        Raises:
            APIError: If token is not returned.

        Returns:
            str: Bearer Token value.
        """
        url = "https://auth.EU.customergauge.com/oauth2/token"
        client_id = self.credentials.get("client_id", None)
        client_secret = self.credentials.get("client_secret", None)

        headers = {"Content-Type": "application/x-www-form-urlencoded"}
        body = {
            "grant_type": "client_credentials",
            "client_id": client_id,
            "client_secret": client_secret,
        }
        api_response = handle_api_response(
            url=url, params=body, headers=headers, method="POST"
        )
        token = api_response.json().get("access_token")

        if token is None:
            raise APIError("The token could not be generated. Check your credentials.")

        return token

    def get_json_response(
        self,
        cursor: int = None,
        pagesize: int = 1000,
        date_field: Literal[
            "date_creation", "date_order", "date_sent", "date_survey_response"
        ] = None,
        start_date: datetime = None,
        end_date: datetime = None,
    ) -> Dict[str, Any]:
        """
        Gets JSON with nested structure that contains data and cursor parameter value using GET request method.

        Args:
            cursor (int, optional): Cursor value to navigate to the page. Defaults to None.
            pagesize (int, optional): Number of responses (records) returned per page, max value = 1000. Defaults to 1000. Defaults to 1000.
            date_field (Literal["date_creation", "date_order", "date_sent", "date_survey_response"], optional): Specifies the date type which filter date range. Defaults to None.
            start_date (datetime, optional): Defines the period start date in yyyy-mm-dd format. Defaults to None.
            end_date (datetime, optional): Defines the period end date in yyyy-mm-dd format. Defaults to None.

        Raises:
            ValueError: If at least one date argument were provided and the rest is missing. Needed all 3 or skip them.
            APIError: If no response from API call.

        Returns:
            Dict[str, Any]: JSON with data and cursor parameter value.
        """
        url = self.url

        params = {
            "per_page": pagesize,
            "with[]": ["drivers", "tags", "questions", "properties"],
            "cursor": cursor,
        }

        if any([date_field, start_date, end_date]):
            if all([date_field, start_date, end_date]):
                params["period[field]"] = (date_field,)
                params["period[start]"] = (start_date,)
                params["period[end]"] = end_date
            else:
                raise ValueError(
                    "Missing date arguments: 'date_field', 'start_date', 'end_date'. Provide all 3 arguments or skip all of them."
                )

        header = {"Authorization": f"Bearer {self.get_token()}"}
        api_response = handle_api_response(url=url, headers=header, params=params)
        response = api_response.json()

        if response is None:
            raise APIError("No response.")
        return response

    def get_cursor(self, json_response: Dict[str, Any] = None) -> int:
        """
        Returns cursor value that is needed to navigate to the next page in the next API call for specific pagesize.

        Args:
            json_response (Dict[str, Any], optional): Dictionary with nested structure that contains data and cursor parameter value. Defaults to None.

        Raises:
            ValueError: If cursor value not found.

        Returns:
            int: Cursor value.
        """

        try:
            cur = json_response["cursor"]["next"]
        except:
            raise ValueError(
                "Provided argument doesn't contain 'cursor' value. Pass json returned from the endpoint."
            )

        return cur

    def properties_cleaning(
        self, json_response: Dict[str, Any] = None
    ) -> Dict[str, Any]:
        """
        Returns initialy cleaned data. The cleaning of the additional params is depend on the endpoint.

        Args:
            json_response (Dict[str, Any], optional): Dictionary with nested structure that contains data and cursor parameter value. Defaults to None.

        Returns:
            Dict[str, Any]: Dictionary that contains cleaned data corresponding to one record.
        """
        clean_properties = {
            d["field"]: d["reference"] for d in json_response["properties"]
        }
        json_response["properties"] = clean_properties

        if self.endpoint == "responses":
            json_response["drivers"] = (
                " ".join(map(str, json_response["drivers"]))
                .replace("label", ",")
                .replace(r"{',':", " ")
                .replace(r"'", "")
                .replace("}", "")
                .strip()
                .replace("  ", ",")
            )
            json_response["tags"] = " ".join(map(str, json_response["tags"])).replace(
                "[]", ""
            )
            json_response["questions"] = " ".join(
                map(str, json_response["questions"])
            ).replace("[]", "")
        else:
            pass

        return json_response

    def flatten_json(self, json_response: Dict[str, Any] = None) -> Dict[str, Any]:
        """
        Function that flattens a nested structure of the JSON object into a single-level dictionary.
        Uses a nested `flatten()` function to recursively combine nested keys in the JSON object with '_' to create the flattened keys.

        Args:
            json_response (Dict[str, Any], optional): JSON object represented as a nested dictionary. Defaults to None.

        Returns:
            Dict[str, Any]: The flattened dictionary.
        """
        out = {}

        def flattify(x, key=""):
            if type(x) is dict:
                for a in x:
                    flattify(x[a], key + a + "_")
            else:
                out[key[:-1]] = x

        flattify(json_response)

        return out

    def to_df(self, json_response: Dict[str, Any] = None) -> pd.DataFrame:
        """
        Flatten dictionary structure and convert it into pandas DataFrame. Cleans column names.

        Args:
            json_response (Dict[str, Any], optional): JSON object represented as a nested dictionary that contains data and cursor parameter value. Defaults to None.

        Raises:
            ValueError: If data value not found.

        Returns:
            pd.DataFrame: pandas.DataFrame
        """
        try:
            response_json = json_response["data"]
        except:
            raise ValueError(
                "Provided argument doesn't contain 'data' value. Pass json returned from the endpoint."
            )
        clean_json = list(map(self.properties_cleaning, response_json))
        df = pd.DataFrame(list(map(self.flatten_json, clean_json)))
        df.columns = df.columns.str.lower().str.replace(" ", "_")

        return df



import os
from pathlib import Path
from typing import Any, Dict, List

import pendulum
from prefect import Flow
from prefect.backend import set_key_value
from prefect.utilities import logging

from viadot.task_utils import (
    add_ingestion_metadata_task,
    df_clean_column,
    df_get_data_types_task,
    df_map_mixed_dtypes_for_parquet,
    df_to_csv,
    df_to_parquet,
    dtypes_to_json_task,
    update_dtypes_dict,
)
from viadot.tasks import AzureDataLakeUpload, SalesforceToDF

logger = logging.get_logger(__name__)


class SalesforceToADLS(Flow):
    def __init__(
        self,
        name: str = None,
        query: str = None,
        table: str = None,
        columns: List[str] = None,
        domain: str = "test",
        client_id: str = "viadot",
        env: str = "DEV",
        vault_name: str = None,
        credentials_secret: str = None,
        output_file_extension: str = ".parquet",
        overwrite_adls: bool = True,
        adls_dir_path: str = None,
        local_file_path: str = None,
        adls_file_name: str = None,
        adls_sp_credentials_secret: str = None,
        if_exists: str = "replace",
        timeout: int = 3600,
        *args: List[Any],
        **kwargs: Dict[str, Any],
    ):
        """
        Flow for downloading data from Salesforce table to a local CSV or Parquet file, then uploading it to Azure Data Lake.

        Args:
            name (str, optional): Flow name. Defaults to None.
            query (str, optional): Query for download the data if specific download is needed. Defaults to None.
            table (str, optional): Table name. Can be used instead of query. Defaults to None.
            columns (List[str], optional): List of columns which are needed - table argument is needed. Defaults to None.
            domain (str, optional): Domain of a connection; defaults to 'test' (sandbox).
                Can only be added if built-in username/password/security token is provided. Defaults to None.
            client_id (str, optional): Client id to keep the track of API calls. Defaults to None.
            env (str, optional): Environment information, provides information about credential
                and connection configuration. Defaults to 'DEV'.
            credentials_secret (str, optional): The name of the Azure Key Vault secret for Salesforce. Defaults to None.
            vault_name (str, optional): The name of the vault from which to obtain the secrets. Defaults to None.
            output_file_extension (str, optional): Output file extension - to allow selection of CSV for data
                which is not easy to handle with parquet. Defaults to ".parquet".
            overwrite_adls (bool, optional): Whether to overwrite the file in ADLS. Defaults to True.
            adls_dir_path (str, optional): Azure Data Lake destination folder/catalog path. Defaults to None.
            local_file_path (str, optional): Local destination path. Defaults to None.
            adls_file_name (str, optional): Name of file in ADLS. Defaults to None.
            adls_sp_credentials_secret (str, optional): The name of the Azure Key Vault secret containing a dictionary with
                ACCOUNT_NAME and Service Principal credentials (TENANT_ID, CLIENT_ID, CLIENT_SECRET) for the Azure Data Lake.
                Defaults to None.
            if_exists (str, optional): What to do if the file exists. Defaults to "replace".
            timeout(int, optional): The amount of time (in seconds) to wait while running this task before
                a timeout occurs. Defaults to 3600.
        """
        # SalesforceToDF
        self.query = query
        self.table = table
        self.columns = columns
        self.domain = domain
        self.client_id = client_id
        self.env = env
        self.vault_name = vault_name
        self.credentials_secret = credentials_secret

        # AzureDataLakeUpload
        self.adls_sp_credentials_secret = adls_sp_credentials_secret
        self.if_exists = if_exists
        self.output_file_extension = output_file_extension
        self.now = str(pendulum.now("utc"))
        self.timeout = timeout

        self.local_file_path = (
            local_file_path or self.slugify(name) + self.output_file_extension
        )
        self.local_json_path = self.slugify(name) + ".json"
        self.adls_dir_path = adls_dir_path

        if adls_file_name is not None:
            self.adls_file_path = os.path.join(adls_dir_path, adls_file_name)
            self.adls_schema_file_dir_file = os.path.join(
                adls_dir_path, "schema", Path(adls_file_name).stem + ".json"
            )
        else:
            self.adls_file_path = os.path.join(
                adls_dir_path, self.now + self.output_file_extension
            )
            self.adls_schema_file_dir_file = os.path.join(
                adls_dir_path, "schema", self.now + ".json"
            )
        self.overwrite_adls = overwrite_adls

        super().__init__(*args, name=name, **kwargs)

        self.gen_flow()

    @staticmethod
    def slugify(name):
        return name.replace(" ", "_").lower()

    def gen_flow(self) -> Flow:
        salesforce_to_df_task = SalesforceToDF(timeout=self.timeout)
        df = salesforce_to_df_task.bind(
            query=self.query,
            table=self.table,
            columns=self.columns,
            domain=self.domain,
            client_id=self.client_id,
            env=self.env,
            vault_name=self.vault_name,
            credentials_secret=self.credentials_secret,
            flow=self,
        )

        df_clean = df_clean_column.bind(df=df, flow=self)
        df_with_metadata = add_ingestion_metadata_task.bind(df_clean, flow=self)
        dtypes_dict = df_get_data_types_task.bind(df_with_metadata, flow=self)
        df_to_be_loaded = df_map_mixed_dtypes_for_parquet(
            df_with_metadata, dtypes_dict, flow=self
        )

        if self.output_file_extension == ".parquet":
            df_to_file = df_to_parquet.bind(
                df=df_to_be_loaded,
                path=self.local_file_path,
                if_exists=self.if_exists,
                flow=self,
            )
        else:
            df_to_file = df_to_csv.bind(
                df=df_with_metadata,
                path=self.local_file_path,
                if_exists=self.if_exists,
                flow=self,
            )

        file_to_adls_task = AzureDataLakeUpload(timeout=self.timeout)
        file_to_adls_task.bind(
            from_path=self.local_file_path,
            to_path=self.adls_file_path,
            overwrite=self.overwrite_adls,
            sp_credentials_secret=self.adls_sp_credentials_secret,
            flow=self,
        )

        dtypes_updated = update_dtypes_dict(dtypes_dict, flow=self)
        dtypes_to_json_task.bind(
            dtypes_dict=dtypes_updated, local_json_path=self.local_json_path, flow=self
        )

        json_to_adls_task = AzureDataLakeUpload(timeout=self.timeout)
        json_to_adls_task.bind(
            from_path=self.local_json_path,
            to_path=self.adls_schema_file_dir_file,
            overwrite=self.overwrite_adls,
            sp_credentials_secret=self.adls_sp_credentials_secret,
            flow=self,
        )

        df_clean.set_upstream(df, flow=self)
        df_with_metadata.set_upstream(df_clean, flow=self)
        dtypes_dict.set_upstream(df_with_metadata, flow=self)
        df_to_be_loaded.set_upstream(dtypes_dict, flow=self)
        file_to_adls_task.set_upstream(df_to_file, flow=self)
        json_to_adls_task.set_upstream(dtypes_to_json_task, flow=self)
        set_key_value(key=self.adls_dir_path, value=self.adls_file_path)



import textwrap
from typing import List

from pyrfc import Connection

from viadot.exceptions import CredentialError, ValidationError
from viadot.sources.base import Source


class SAPBW(Source):
    """
    A class for quering the SAP BW (SAP Business Warehouse) source using pyrfc library.
    Documentation to pyrfc can be found under: https://sap.github.io/PyRFC/pyrfc.html
    Documentation for SAP connection modules under: https://www.se80.co.uk/sap-function-modules/list/?index=rsr_mdx

    """

    def __init__(self, credentials: dict, *args, **kwargs):
        """
        Create an instance of the SAPBW class.

        Args:
            credentials (dict): Credentials to connect with SAP BW containing ashost, sysnr, user, passwd, client.

        Raises:
            CredentialError: If provided credentials are incorrect.
        """
        if credentials is None:
            raise CredentialError("Missing credentials.")

        super().__init__(*args, credentials=credentials, **kwargs)

    def get_connection(self) -> Connection:
        """
        Function to create the connection with SAP BW.

        Returns:
            Connection: Connection to SAP.
        """
        return Connection(
            ashost=self.credentials.get("ashost"),
            sysnr=self.credentials.get("sysnr"),
            user=self.credentials.get("user"),
            passwd=self.credentials.get("passwd"),
            client=self.credentials.get("client"),
        )

    def get_all_available_columns(self, mdx_query: str) -> List:
        """
        Function to generate list of all available columns in the SAP table based on passed MDX query.

        Args:
            mdx_query (str): The MDX query to be passed to connection.

        Returns:
            all_available_columns: List of all available columns in the source table.
        """

        conn = self.get_connection()
        query = textwrap.wrap(
            mdx_query, width=75
        )  # width = 75, to properly split mdx query into substrings passed to SAP object creation function

        properties = conn.call("RSR_MDX_CREATE_STORED_OBJECT", COMMAND_TEXT=query)
        datasetid = properties["DATASETID"]

        if properties["RETURN"]["MESSAGE"] == "":
            get_axis_info = conn.call(
                "RSR_MDX_GET_AXIS_INFO", DATASETID=datasetid
            )  # listing all of available columns and metrics
            cols = get_axis_info["AXIS_DIMENSIONS"]

            all_available_columns = [x["DIM_UNAM"] for x in cols]
        else:
            all_available_columns = []
            self.logger.error(properties["RETURN"]["MESSAGE"])

        return all_available_columns

    def get_output_data(self, mdx_query: str) -> dict:
        """
        Function to generate the SAP output dataset from MDX query.

        Args:
            mdx_query (str): The MDX query to be passed to connection.

        Returns:
            query_output: SAP output dictionary in JSON format that contains data rows and column headers.

            Example output:

                {
                "RETURN": {
                    "TYPE": "",
                    "ID": "",
                    "NUMBER": "000",
                    "MESSAGE": "",...
                },...
                "DATA": [
                    {
                    "COLUMN": 0,
                    "ROW": 0,
                    "DATA": "VELUX Deutschland GmbH",
                    "VALUE_DATA_TYPE": "CHAR",
                    "CELL_STATUS": ""
                    },...
                ],
                "HEADER": [
                    {
                    "COLUMN": 0,
                    "ROW": 0,
                    "DATA": "[0COMP_CODE].[LEVEL01].[DESCRIPTION]",
                    "VALUE_DATA_TYPE": "CHAR",
                    "CELL_STATUS": ""
                    },...
                ]
                }

        """
        conn = self.get_connection()
        query = textwrap.wrap(
            mdx_query, 75
        )  # width = 75, to properly split mdx query into substrings passed to SAP object creation function
        properties = conn.call("RSR_MDX_CREATE_OBJECT", COMMAND_TEXT=query)

        datasetid = properties["DATASETID"]
        query_output = conn.call("RSR_MDX_GET_FLAT_DATA", DATASETID=datasetid)

        return query_output



from prefect.tasks.secrets import PrefectSecret

from viadot.config import local_config
from viadot.tasks import C4CReportToDF, C4CToDF


def test_c4c_to_df():
    url = "http://services.odata.org/V2/Northwind/Northwind.svc/"
    endpoint = "Employees"
    c4c_to_df = C4CToDF()
    df = c4c_to_df.run(url=url, endpoint=endpoint)
    answer = df.head()
    assert answer.shape[1] == 23


def test_c4c_report_to_df():
    credentials = local_config.get("CLOUD_FOR_CUSTOMERS")
    credentials_prod = credentials["Prod"]
    report_url = credentials_prod["server"]
    c4c_report_to_df = C4CReportToDF()
    df = c4c_report_to_df.run(report_url=report_url, env="Prod")
    answer = df.head()

    assert answer.shape[0] == 5


def test_c4c_to_df_kv():
    task = C4CToDF()
    credentials_secret = PrefectSecret("C4C_KV").run()
    res = task.run(credentials_secret=credentials_secret, endpoint="ActivityCollection")
    answer = res.head()
    assert answer.shape[1] == 19



import pandas as pd

from viadot.tasks.mediatool import MediatoolToDF


def test_join_df():
    data_left = {"id": [102, 112], "amount": [33, 48]}
    df_left = pd.DataFrame.from_dict(data=data_left)
    data_right = {"id_r": [103, 112], "prize": [12, 234], "name": ["cat", "dog"]}
    df_right = pd.DataFrame.from_dict(data=data_right)

    mtool = MediatoolToDF(mediatool_credentials_key="MEDIATOOL-TESTS")

    df = mtool.join_dfs(
        df_left=df_left,
        df_right=df_right,
        left_on="id",
        right_on="id_r",
        columns_from_right_df=["id_r", "prize"],
        how="left",
    )

    assert isinstance(df, pd.DataFrame)
    assert "prize" in df.columns
    assert not df.empty



import os
from unittest import mock

import pandas as pd
import pytest

from viadot.flows import EurostatToADLS

DATA = {"geo": ["PL", "DE", "NL"], "indicator": [35, 55, 77]}
ADLS_FILE_NAME = "test_eurostat.parquet"
ADLS_DIR_PATH = "raw/tests/"


@mock.patch(
    "viadot.tasks.EurostatToDF.run",
    return_value=pd.DataFrame(data=DATA),
)
@pytest.mark.run
def test_eurostat_to_adls_run_flow(mocked_class):
    flow = EurostatToADLS(
        "test_eurostat_to_adls_flow_run",
        dataset_code="ILC_DI04",
        overwrite_adls=True,
        adls_dir_path=ADLS_DIR_PATH,
        adls_file_name=ADLS_FILE_NAME,
    )
    result = flow.run()
    assert result.is_successful()
    os.remove("test_eurostat_to_adls_flow_run.parquet")
    os.remove("test_eurostat_to_adls_flow_run.json")



from viadot.tasks import BlobFromCSV


def test_blob_from_csv(TEST_CSV_FILE_PATH, TEST_CSV_FILE_BLOB_PATH):
    blob_task = BlobFromCSV()
    result = blob_task.run(
        from_path=TEST_CSV_FILE_PATH, to_path=TEST_CSV_FILE_BLOB_PATH, overwrite=True
    )
    assert result == True



import pandas as pd
import pytest

from viadot.tasks import SalesforceToDF, SalesforceUpsert


@pytest.fixture(scope="session")
def test_df():
    data = {
        "Id": ["111"],
        "LastName": ["John Tester-External 3"],
        "SAPContactId__c": [111],
    }
    df = pd.DataFrame(data=data)
    yield df


def test_salesforce_upsert(test_df):
    """
    Id and SAPContactId__c are unique values, you can update only non-unique values for this test.
    If the combiantion of Id and SAPContactId__c do not exist, the test will fail.
    The Id and SAPContactId__c values '111' needs to be replaced with proper one (that exist in the testing system).
    """
    try:
        sf = SalesforceUpsert()
        sf.run(test_df, table="Contact")
    except Exception as exception:
        assert False, exception


def test_salesforce_to_df():
    sf_to_df = SalesforceToDF(
        query="SELECT IsDeleted, FiscalYear FROM Opportunity LIMIT 50"
    )
    df = sf_to_df.run()

    assert isinstance(df, pd.DataFrame)
    assert (50, 2) == df.shape



from .adls_container_to_container import ADLSContainerToContainer
from .adls_gen1_to_azure_sql import ADLSGen1ToAzureSQL
from .adls_gen1_to_azure_sql_new import ADLSGen1ToAzureSQLNew
from .adls_gen1_to_gen2 import ADLSGen1ToGen2
from .adls_to_azure_sql import ADLSToAzureSQL
from .aselite_to_adls import ASELiteToADLS
from .azure_sql_transform import AzureSQLTransform
from .bigquery_to_adls import BigQueryToADLS
from .cloud_for_customers_report_to_adls import CloudForCustomersReportToADLS
from .flow_of_flows import Pipeline
from .genesys_to_adls import GenesysToADLS
from .outlook_to_adls import OutlookToADLS
from .salesforce_to_adls import SalesforceToADLS
from .sharepoint_to_adls import SharepointToADLS
from .supermetrics_to_adls import SupermetricsToADLS
from .supermetrics_to_azure_sql import SupermetricsToAzureSQL

try:
    from .sap_to_duckdb import SAPToDuckDB
except ImportError:
    pass

from .duckdb_to_sql_server import DuckDBToSQLServer
from .duckdb_transform import DuckDBTransform
from .multiple_flows import MultipleFlows
from .prefect_logs import PrefectLogs

try:
    from .sap_rfc_to_adls import SAPRFCToADLS
except ImportError:
    pass

try:
    from .sap_bw_to_adls import SAPBWToADLS
except ImportError:
    pass

from .customer_gauge_to_adls import CustomerGaugeToADLS
from .epicor_to_duckdb import EpicorOrdersToDuckDB
from .eurostat_to_adls import EurostatToADLS
from .hubspot_to_adls import HubspotToADLS
from .mediatool_to_adls import MediatoolToADLS
from .mindful_to_adls import MindfulToADLS
from .sftp_operations import SftpToADLS, SftpToAzureSQL
from .sql_server_to_duckdb import SQLServerToDuckDB
from .sql_server_to_parquet import SQLServerToParquet
from .sql_server_transform import SQLServerTransform
from .transform_and_catalog import TransformAndCatalogToLuma
from .vid_club_to_adls import VidClubToADLS



import pytest

from viadot.flows import PrefectLogs


@pytest.fixture(scope="session")
def expectation_suite():
    expectation_suite = {
        "data": {
            "project": [
                {
                    "id": "223a8acf-4cf0-4cf7-ae1f-b66f78e28813",
                    "name": "Admin",
                    "flows": [
                        {
                            "id": "b13dcc6d-b621-4acd-88be-2cf28715a7c5",
                            "name": "1-raw table_catalog extract",
                            "version": 3,
                            "flow_runs": [
                                {
                                    "id": "d1d76bbd-b494-4cf2-bf59-3dfccf520039",
                                    "scheduled_start_time": "2022-09-06T09:19:47.937928+00:00",
                                    "start_time": "2022-09-06T09:20:06.944586+00:00",
                                    "end_time": "2022-09-06T09:20:39.386856+00:00",
                                    "state": "Cancelled",
                                    "created_by_user_id": "5878be75-ee66-42f4-8179-997450063ea4",
                                }
                            ],
                        },
                        {
                            "id": "14b1a89e-f902-48a1-b6df-43cacdb91e1a",
                            "name": "1-raw table_catalog extract",
                            "version": 2,
                            "flow_runs": [],
                        },
                        {
                            "id": "a1eace09-38b4-46bf-bacf-a5d29bdbb633",
                            "name": "1-raw table_catalog extract",
                            "version": 1,
                            "flow_runs": [],
                        },
                    ],
                },
                {
                    "id": "512d0f29-2ceb-4177-b7d8-c5908da666ef",
                    "name": "integrations",
                    "flows": [],
                },
                {
                    "id": "1d3c5246-61e5-4aff-a07b-4b74959a46e4",
                    "name": "dev_cdl",
                    "flows": [],
                },
                {
                    "id": "eac9b6d4-725a-4354-bf8f-25e7828ea2d8",
                    "name": "Admin",
                    "flows": [],
                },
                {
                    "id": "52217a3c-f42a-4448-afc4-2a325415b8e8",
                    "name": "test",
                    "flows": [],
                },
                {
                    "id": "516b3fe9-1c26-47a4-b797-e6a77bee390c",
                    "name": "cic",
                    "flows": [],
                },
                {
                    "id": "ebe0e5aa-4add-4440-8c1a-6f9c74eb29fe",
                    "name": "dev",
                    "flows": [],
                },
            ]
        }
    }

    yield expectation_suite


def test_prefect_logs(expectation_suite):
    flow = PrefectLogs(
        name="Extract prefect data test",
        query="""
                {
                        project {
                            id
                            name
                            flows (
                                where : {name: {_eq: "1-raw table_catalog extract"}}
                            ) {
                                    id
                                    name
                                    version
                                    flow_runs(
                                        order_by: {end_time: desc}
                                        where: {_and:
                                            [
                                            {scheduled_start_time:{ %s: "%s" }},
                                            {state: {_neq: "Scheduled"}}
                                            ]
                                        }
                                        )
                                            {
                                            id
                                            scheduled_start_time
                                            start_time
                                            end_time
                                            state
                                            created_by_user_id
                                            }
                            }
                        }
                    }
        """,
        scheduled_start_time="2022-09-05",
        filter_type="_gte",
        local_file_path=f"prefect_extract_logs.parquet",
        adls_path=f"raw/tests/prefect_extract_logs.parquet",
    )

    results = flow.run()
    assert results.is_successful()



import json
from prefect.tasks.shell import ShellTask
from .azure_key_vault import AzureKeyVaultSecret


class LumaIngest(ShellTask):
    """Ingest metadata into [Luma](https://github.com/dyvenia/luma)."""

    def __init__(
        self,
        metadata_dir_path: str,
        url: str = None,
        dbt_project_path: str = None,
        credentials_secret: str = None,
        vault_name: str = None,
        *args,
        **kwargs,
    ) -> None:
        """
        Runs Luma ingestion by sending dbt artifacts to Luma ingestion API.

        Args:
            metadata_dir_path (str): The path to the directory containing metadata files.
                In the case of dbt, it's dbt project's `target` directory, which contains dbt artifacts
                (`sources.json`, `catalog.json`, `manifest.json`, and `run_results.json`).
            url (str, optional): The url of the Luma ingestion API. Defaults to None.
            credentials_secret (str, optional): The name of the Azure Key Vault secret containing Luma credentials.
                Defaults to None.
            vault_name (str, optional): The name of the vault from which to obtain the secrets. Defaults to None.
        """
        super().__init__(*args, **kwargs)

        if not url:
            azure_secret_task = AzureKeyVaultSecret()
            credentials_str = azure_secret_task.run(
                secret=credentials_secret, vault_name=vault_name
            )
            url = json.loads(credentials_str).get("url")
        self.helper_script = dbt_project_path
        self.url = url
        self.metadata_dir_path = metadata_dir_path
        self.command = f"luma dbt send-test-results --luma-url {url} --metadata-dir {metadata_dir_path}"
        self.return_all = True
        self.stream_output = True
        self.log_stderr = True



import io
from unittest import mock

import pandas as pd
import pytest
from pytest import fixture, raises

from viadot.sources.sftp import SftpConnector
from viadot.tasks.sftp import SftpList, SftpToDF


@pytest.fixture
def tmp_df():
    data = {"country": [1, 2], "sales": [3, 4]}
    df = pd.DataFrame(data=data)
    return df


@pytest.fixture
def list_of_paths():
    list_of_paths = [
        "Country__Context30##exported.tsv",
        "Country__Context31##exported.tsv",
        "Country__Context32##exported.tsv",
        "Country__Context33##exported.tsv",
        "Country__Context4##exported.tsv",
        "Country__Context6##exported.tsv",
        "Country__Context7##exported.tsv",
        "Country__Context8##exported.tsv",
        "Local_Products.csv",
        "Products Checkup.tsv",
        "Products.tsv",
        "RewardTest.csv",
    ]
    return list_of_paths


@pytest.fixture
def df_buf():
    s_buf = io.StringIO()

    data = {"country": [1, 2], "sales": [3, 4]}
    df = pd.DataFrame(data=data)
    df.to_csv(s_buf)
    return s_buf


def test_create_sftp_instance():
    s = SftpConnector(
        credentials_sftp={"HOSTNAME": 1, "USERNAME": 2, "PASSWORD": 3, "PORT": 4}
    )
    assert s


def test_connection_sftp(tmp_df):
    with mock.patch("viadot.sources.sftp.SftpConnector.to_df") as mock_method:
        mock_method.return_value = tmp_df
        s = SftpConnector(
            credentials_sftp={"HOSTNAME": 1, "USERNAME": 2, "PASSWORD": 3, "PORT": 4}
        )

        final_df = s.to_df()
        assert isinstance(final_df, pd.DataFrame)


def test_getfo_file(df_buf):
    with mock.patch("viadot.sources.sftp.SftpConnector.getfo_file") as mock_method:
        mock_method.return_value = df_buf
        s = SftpConnector(
            credentials_sftp={"HOSTNAME": 1, "USERNAME": 2, "PASSWORD": 3, "PORT": 4}
        )

        buffer_df = s.getfo_file()
        buffer_df.seek(0)
        df = pd.read_csv(buffer_df)
        assert isinstance(df, pd.DataFrame)


def test_ls_sftp(list_of_paths):
    with mock.patch("viadot.sources.sftp.SftpConnector.list_directory") as mock_method:
        mock_method.return_value = list_of_paths
        s = SftpConnector(
            credentials_sftp={"HOSTNAME": 1, "USERNAME": 2, "PASSWORD": 3, "PORT": 4}
        )

        paths = s.list_directory()

        assert isinstance(paths, list)


def test_get_exported_files(list_of_paths):
    with mock.patch.object(
        SftpConnector, "get_exported_files", return_value=list_of_paths
    ) as mock_method:
        s = SftpConnector(
            credentials_sftp={"HOSTNAME": 1, "USERNAME": 2, "PASSWORD": 3, "PORT": 4}
        )
        filtered_paths = s.get_exported_files()

        assert isinstance(filtered_paths, list)


def test_task_sftptodf(tmp_df):
    task = SftpToDF()
    with mock.patch.object(SftpToDF, "run", return_value=tmp_df) as mock_method:
        df = task.run()

        assert isinstance(df, pd.DataFrame)


def test_task_sftplist(list_of_paths):
    task = SftpList()
    with mock.patch.object(SftpList, "run", return_value=list_of_paths) as mock_method:
        list_directory = task.run()

        assert isinstance(list_directory, list)


def test_example():
    with mock.patch("viadot.sources.sftp.SftpConnector.to_df") as mock_method:
        mock_method.return_value = tmp_df
        t = SftpToDF()
        t.run()



from unittest import mock

import pandas as pd
import pytest

from viadot.task_utils import credentials_loader
from viadot.tasks import VidClubToDF

CREDENTIALS = credentials_loader.run(credentials_secret="VIDCLUB")


class MockVidClubResponse:
    response_data = pd.DataFrame()


@pytest.fixture(scope="session")
def var_dictionary():
    variables = {
        "source": "jobs",
        "from_date": "2022-03-23",
        "to_date": "2022-03-24",
        "items_per_page": 1,
        "days_interval": 1,
    }
    yield variables


@mock.patch(
    "viadot.tasks.VidClubToDF.run", return_value=MockVidClubResponse.response_data
)
def test_vid_club_to_df(var_dictionary):
    """
    Checks if run method returns DataFrame.

    Args:
        var_dictionary: Dictionary with example arguments for run method.
    """
    vc_to_df = VidClubToDF(credentials=CREDENTIALS)

    df = vc_to_df.run(
        source=var_dictionary["source"],
        to_date=var_dictionary["to_date"],
        from_date=var_dictionary["from_date"],
        items_per_page=var_dictionary["items_per_page"],
        days_interval=var_dictionary["days_interval"],
    )

    assert isinstance(df, pd.DataFrame)


@pytest.mark.drop_cols
def test_drop_columns(var_dictionary):
    """
    Tests cols_to_drop argument in function.

    Args:
        var_dictionary: Dictionary with example arguments for run method.
    """
    cols_to_drop = ["regionID", "submissionDate"]
    vc_to_df = VidClubToDF(credentials=CREDENTIALS)

    output_with_dropped = vc_to_df.run(
        source=var_dictionary["source"],
        to_date=var_dictionary["to_date"],
        from_date=var_dictionary["from_date"],
        items_per_page=var_dictionary["items_per_page"],
        days_interval=var_dictionary["days_interval"],
        cols_to_drop=cols_to_drop,
    )

    assert all(col not in output_with_dropped.columns for col in cols_to_drop)


@pytest.mark.drop_cols
def test_drop_columns_KeyError(var_dictionary, caplog):
    """
    Tests if in case of KeyError (when passed columns in cols_to_drop are not included in DataFrame), there is returned error logger..

    Args:
        var_dictionary: Dictionary with example arguments for run method.
    """
    cols_to_drop = ["Test", "submissionDate"]
    vc_to_df = VidClubToDF(credentials=CREDENTIALS)

    vc_to_df.run(
        source=var_dictionary["source"],
        to_date=var_dictionary["to_date"],
        from_date=var_dictionary["from_date"],
        items_per_page=var_dictionary["items_per_page"],
        days_interval=var_dictionary["days_interval"],
        cols_to_drop=cols_to_drop,
    )
    assert len(caplog.records) == 1
    assert caplog.records[0].levelname == "ERROR"
    assert (
        f"Column(s): {cols_to_drop} don't exist in the DataFrame"
        in caplog.records[0].message
    )


@pytest.mark.drop_cols
def test_drop_columns_TypeError(var_dictionary):
    """
    Tests raising TypeError if passed columns in cols_to_drop is not a List.

    Args:
        var_dictionary: Dictionary with example arguments for run method.
    """
    with pytest.raises(TypeError, match="Provide columns to drop in a List."):
        cols_to_drop = "Test"
        vc_to_df = VidClubToDF(credentials=CREDENTIALS)

        output_with_dropped = vc_to_df.run(
            source=var_dictionary["source"],
            to_date=var_dictionary["to_date"],
            from_date=var_dictionary["from_date"],
            items_per_page=var_dictionary["items_per_page"],
            days_interval=var_dictionary["days_interval"],
            cols_to_drop=cols_to_drop,
        )



import pandas as pd
import pytest

from viadot.task_utils import credentials_loader
from viadot.tasks import SAPBWToDF

CREDENTIALS = credentials_loader.run(credentials_secret="SAP")
sapbw_task = SAPBWToDF(sapbw_credentials=CREDENTIALS.get("BW"))


@pytest.fixture(scope="session")
def output_variable():
    output = {
        "RETURN": {
            "TYPE": "",
            "ID": "",
            "NUMBER": "000",
            "MESSAGE": "",
            "LOG_NO": "",
            "LOG_MSG_NO": "000000",
            "MESSAGE_V1": "",
            "MESSAGE_V2": "",
            "MESSAGE_V3": "",
            "MESSAGE_V4": "",
            "PARAMETER": "",
            "ROW": 0,
            "FIELD": "",
            "SYSTEM": "",
        },
        "STATISTIC": {"STEP": "003YPR44RQTVS3BSMZTKDYBMD"},
        "DATA": [
            {
                "COLUMN": 0,
                "ROW": 0,
                "DATA": "January 2023",
                "VALUE_DATA_TYPE": "CHAR",
                "CELL_STATUS": "",
            },
            {
                "COLUMN": 1,
                "ROW": 0,
                "DATA": "202301",
                "VALUE_DATA_TYPE": "NUMC",
                "CELL_STATUS": "",
            },
        ],
        "HEADER": [
            {
                "COLUMN": 0,
                "ROW": 0,
                "DATA": "[0CALMONTH].[LEVEL01].[DESCRIPTION]",
                "VALUE_DATA_TYPE": "CHAR",
                "CELL_STATUS": "",
            },
            {
                "COLUMN": 1,
                "ROW": 0,
                "DATA": "[0CALMONTH].[LEVEL01].[MEMBER_NAME]",
                "VALUE_DATA_TYPE": "CHAR",
                "CELL_STATUS": "",
            },
        ],
    }
    yield output


@pytest.fixture(scope="session")
def user_mapping():
    mapping = {
        "[0CALMONTH].[LEVEL01].[DESCRIPTION]": "Calendar Year/Month",
        "[0CALMONTH].[LEVEL01].[MEMBER_NAME]": "Calendar Year/Month key",
    }
    yield mapping


@pytest.fixture(scope="session")
def mdx_query_variable():
    mdx_query = """
        SELECT
                {
            }
                ON COLUMNS,
        NON EMPTY
                { 
                    { [0CALMONTH].[202301] } 
        } 
        DIMENSION PROPERTIES
        DESCRIPTION,
        MEMBER_NAME
        ON ROWS

        FROM ZCSALORD1/ZBW4_ZCSALORD1_006_BOA
    
                """
    yield mdx_query


df_to_test = pd.DataFrame(
    data={
        "[0CALMONTH].[LEVEL01].[DESCRIPTION]": ["January 2023"],
        "[0CALMONTH].[LEVEL01].[MEMBER_NAME]": ["202301"],
        "date": ["2023-06-19 11:12:43+00:00"],
    },
)


def test_apply_user_mapping(user_mapping):
    apply_mapping = sapbw_task.apply_user_mapping(df_to_test, user_mapping)
    print(user_mapping.values())
    assert list(apply_mapping.columns) == list(user_mapping.values())
    assert isinstance(apply_mapping, pd.DataFrame)


def test_to_df(output_variable):
    df = sapbw_task.to_df(output_variable)
    assert isinstance(df, pd.DataFrame)


def test_run(mdx_query_variable, user_mapping):
    df = sapbw_task.run(mdx_query_variable, user_mapping)
    assert isinstance(df, pd.DataFrame)



from datetime import timedelta
from typing import Any, Dict, Literal

from prefect import Task
from prefect.utilities.tasks import defaults_from_attrs

from viadot.sources import sql_server

from ..config import local_config
from ..sources import SQLServer


class SQLServerCreateTable(Task):
    """
    Create a table in SQL Server.

    Args:
        schema (str, optional): Destination schema.
        table (str, optional): Destination table.
        dtypes (Dict[str, Any], optional): Data types to enforce.
        if_exists (Literal, optional): What to do if the table already exists.
        credentials (dict, optional): Credentials for the connection.
        timeout(int, optional): The amount of time (in seconds) to wait while running this task before
            a timeout occurs. Defaults to 3600.
    """

    def __init__(
        self,
        schema: str = None,
        table: str = None,
        dtypes: Dict[str, Any] = None,
        if_exists: Literal["fail", "replace", "skip", "delete"] = "fail",
        credentials: dict = None,
        max_retries: int = 3,
        retry_delay: timedelta = timedelta(seconds=10),
        timeout: int = 3600,
        *args,
        **kwargs,
    ):
        self.schema = schema
        self.table = table
        self.dtypes = dtypes
        self.if_exists = if_exists
        self.credentials = credentials
        super().__init__(
            name="sql_server_create_table",
            max_retries=max_retries,
            retry_delay=retry_delay,
            timeout=timeout,
            *args,
            **kwargs,
        )

    @defaults_from_attrs("if_exists")
    def run(
        self,
        schema: str = None,
        table: str = None,
        dtypes: Dict[str, Any] = None,
        if_exists: Literal["fail", "replace", "skip", "delete"] = None,
        credentials: str = None,
        max_retries: int = None,
        retry_delay: timedelta = None,
    ):
        """
        Create a table in SQL Server.

        Args:
            schema (str, optional): Destination schema.
            table (str, optional): Destination table.
            dtypes (Dict[str, Any], optional): Data types to enforce.
            if_exists (Literal, optional): What to do if the table already exists.
            credentials (dict, optional): Credentials for the connection.
        """

        if credentials is None:
            credentials = local_config.get("SQL_SERVER").get("DEV")
        sql_server = SQLServer(credentials=credentials)

        fqn = f"{schema}.{table}" if schema is not None else table
        created = sql_server.create_table(
            schema=schema, table=table, dtypes=dtypes, if_exists=if_exists
        )
        if created:
            self.logger.info(f"Successfully created table {fqn}.")
        else:
            self.logger.info(
                f"Table {fqn} has not been created as if_exists is set to {if_exists}."
            )


class SQLServerToDF(Task):
    def __init__(
        self,
        config_key: str = None,
        timeout: int = 3600,
        *args,
        **kwargs,
    ):
        """
        Task for downloading data from SQL Server to a pandas DataFrame.

        Args:
            config_key (str, optional): The key inside local config containing the credentials. Defaults to None.
            timeout(int, optional): The amount of time (in seconds) to wait while running this task before
                a timeout occurs. Defaults to 3600.

        """
        self.config_key = config_key

        super().__init__(name="sql_server_to_df", timeout=timeout, *args, **kwargs)

    @defaults_from_attrs("config_key")
    def run(
        self,
        query: str,
        config_key: str = None,
    ):
        """
        Load the result of a SQL Server Database query into a pandas DataFrame.

        Args:
            query (str, required): The query to execute on the SQL Server database. If the qery doesn't start
                with "SELECT" returns an empty DataFrame.
            config_key (str, optional): The key inside local config containing the credentials. Defaults to None.

        """
        if config_key is None:
            config_key = "SQL_SERVER"
        sql_server = SQLServer(config_key=config_key)
        df = sql_server.to_df(query=query)
        nrows = df.shape[0]
        ncols = df.shape[1]

        self.logger.info(
            f"Successfully downloaded {nrows} rows and {ncols} columns of data to a DataFrame."
        )
        return df


class SQLServerQuery(Task):
    def __init__(
        self,
        config_key: str = None,
        timeout: int = 3600,
        *args,
        **kwargs,
    ):
        """
        Task for running queries on SQL Server.

        Args:
            config_key (str, optional): The key inside local config containing the credentials. Defaults to None.
            timeout(int, optional): The amount of time (in seconds) to wait while running this task before
                a timeout occurs. Defaults to 3600.
        """
        self.config_key = config_key

        super().__init__(name="sql_server_query", timeout=timeout, *args, **kwargs)

    @defaults_from_attrs("config_key")
    def run(
        self,
        query: str,
        config_key: str = None,
    ):
        """
        Ran query on SQL Server.

        Args:
            query (str, required): The query to execute on the SQL Server database.
            config_key (str, optional): The key inside local config containing the credentials. Defaults to None.

        """
        if config_key is None:
            config_key = "SQL_SERVER"
        sql_server = SQLServer(config_key=config_key)
        result = sql_server.run(query)

        self.logger.info(f"Successfully ran the query.")
        return result



import logging
import os

import pandas as pd
import pytest

from viadot.signals import SKIP
from viadot.utils import (
    check_if_empty_file,
    gen_bulk_insert_query_from_df,
    add_viadot_metadata_columns,
)

EMPTY_CSV_PATH = "empty.csv"
EMPTY_PARQUET_PATH = "empty.parquet"


class ClassForDecorator:
    source = "Source_name"

    def __init__(self):
        self.df = pd.DataFrame({"a": [123], "b": ["abc"]})

    def to_df(self):
        return self.df

    @add_viadot_metadata_columns()
    def to_df_decorated(self):
        return self.df

    @add_viadot_metadata_columns(source)
    def to_df_decorated_parameter(self):
        return self.df


def test_single_quotes_inside():
    TEST_VALUE = "a'b"
    df1 = pd.DataFrame(
        {
            "a": [
                TEST_VALUE,
            ],
            "b": ["a"],
        }
    )
    test_insert_query = gen_bulk_insert_query_from_df(
        df1, table_fqn="test_schema.test_table"
    )
    TEST_VALUE_ESCAPED = "'a''b'"
    assert (
        test_insert_query
        == f"""INSERT INTO test_schema.test_table (a, b)

VALUES ({TEST_VALUE_ESCAPED}, 'a')"""
    ), test_insert_query


def test_single_quotes_outside():
    TEST_VALUE = "'a'"
    df1 = pd.DataFrame(
        {
            "a": [
                TEST_VALUE,
            ],
            "b": ["b"],
        }
    )
    test_insert_query = gen_bulk_insert_query_from_df(
        df1, table_fqn="test_schema.test_table"
    )
    TEST_VALUE_ESCAPED = "'''a'''"
    assert (
        test_insert_query
        == f"""INSERT INTO test_schema.test_table (a, b)

VALUES ({TEST_VALUE_ESCAPED}, 'b')"""
    ), test_insert_query


def test_double_quotes_inside():
    TEST_VALUE = 'a "b"'
    df1 = pd.DataFrame(
        {
            "a": [
                TEST_VALUE,
            ],
            "b": ["c"],
        }
    )
    test_insert_query = gen_bulk_insert_query_from_df(
        df1, table_fqn="test_schema.test_table"
    )
    TEST_VALUE_ESCAPED = """'a "b"'"""
    assert (
        test_insert_query
        == f"""INSERT INTO test_schema.test_table (a, b)

VALUES ({TEST_VALUE_ESCAPED}, 'c')"""
    ), test_insert_query


def test_check_if_empty_file_csv(caplog):
    with open(EMPTY_CSV_PATH, "w"):
        pass

    with caplog.at_level(logging.WARNING):
        check_if_empty_file(path=EMPTY_CSV_PATH, if_empty="warn")
        assert f"Input file - '{EMPTY_CSV_PATH}' is empty." in caplog.text
    with pytest.raises(ValueError):
        check_if_empty_file(path=EMPTY_CSV_PATH, if_empty="fail")
    with pytest.raises(SKIP):
        check_if_empty_file(path=EMPTY_CSV_PATH, if_empty="skip")

    os.remove(EMPTY_CSV_PATH)


def test_check_if_empty_file_parquet(caplog):
    with open(EMPTY_PARQUET_PATH, "w"):
        pass

    with caplog.at_level(logging.WARNING):
        check_if_empty_file(path=EMPTY_PARQUET_PATH, if_empty="warn")
        assert f"Input file - '{EMPTY_PARQUET_PATH}' is empty." in caplog.text
    with pytest.raises(ValueError):
        check_if_empty_file(path=EMPTY_PARQUET_PATH, if_empty="fail")
    with pytest.raises(SKIP):
        check_if_empty_file(path=EMPTY_PARQUET_PATH, if_empty="skip")

    os.remove(EMPTY_PARQUET_PATH)


def test_check_if_empty_file_no_data(caplog):
    df = pd.DataFrame({"col1": []})
    df.to_parquet(EMPTY_PARQUET_PATH)
    with caplog.at_level(logging.WARNING):
        check_if_empty_file(path=EMPTY_PARQUET_PATH, if_empty="warn")
        assert f"Input file - '{EMPTY_PARQUET_PATH}' is empty." not in caplog.text


def test_add_viadot_metadata_columns_base():
    df_base = ClassForDecorator().to_df()
    df_decorated = ClassForDecorator().to_df_decorated()

    assert df_base.columns.to_list() == ["a", "b"]
    assert df_decorated.columns.to_list() == ["a", "b", "_viadot_source"]
    assert df_decorated["_viadot_source"][0] == "ClassForDecorator"


def test_add_viadot_metadata_columns_with_parameter():
    df_base = ClassForDecorator().to_df()
    df_decorated = ClassForDecorator().to_df_decorated_parameter()

    assert df_base.columns.to_list() == ["a", "b"]
    assert df_decorated.columns.to_list() == ["a", "b", "_viadot_source"]
    assert df_decorated["_viadot_source"][0] == "Source_name"



import os
from unittest import mock

import pandas as pd
import pytest
from prefect.engine import signals

from viadot.flows import ADLSToAzureSQL
from viadot.flows.adls_to_azure_sql import check_dtypes_sort, df_to_csv_task


def test_get_promoted_adls_path_csv_file():
    adls_path_file = "raw/supermetrics/adls_ga_load_times_fr_test/2021-07-14T13%3A09%3A02.997357%2B00%3A00.csv"
    flow = ADLSToAzureSQL(name="test", adls_path=adls_path_file)
    promoted_path = flow.get_promoted_path(env="conformed")
    assert (
        promoted_path
        == "conformed/supermetrics/adls_ga_load_times_fr_test/2021-07-14T13%3A09%3A02.997357%2B00%3A00.csv"
    )


def test_get_promoted_adls_path_parquet_file():
    adls_path_file = "raw/supermetrics/adls_ga_load_times_fr_test/2021-07-14T13%3A09%3A02.997357%2B00%3A00.parquet"
    flow = ADLSToAzureSQL(name="test", adls_path=adls_path_file)
    promoted_path = flow.get_promoted_path(env="conformed")
    assert promoted_path == "conformed/supermetrics/adls_ga_load_times_fr_test.csv"


def test_get_promoted_adls_path_file_starts_with_slash():
    adls_path_dir_starts_with_slash = "/raw/supermetrics/adls_ga_load_times_fr_test/"
    flow = ADLSToAzureSQL(name="test", adls_path=adls_path_dir_starts_with_slash)
    promoted_path = flow.get_promoted_path(env="conformed")
    assert promoted_path == "conformed/supermetrics/adls_ga_load_times_fr_test.csv"


def test_get_promoted_adls_path_dir_slash():
    adls_path_dir_slash = "raw/supermetrics/adls_ga_load_times_fr_test/"
    flow = ADLSToAzureSQL(name="test", adls_path=adls_path_dir_slash)
    promoted_path = flow.get_promoted_path(env="conformed")
    assert promoted_path == "conformed/supermetrics/adls_ga_load_times_fr_test.csv"


def test_get_promoted_adls_path_dir():
    adls_path_dir = "raw/supermetrics/adls_ga_load_times_fr_test"
    flow = ADLSToAzureSQL(name="test", adls_path=adls_path_dir)
    promoted_path = flow.get_promoted_path(env="conformed")
    assert promoted_path == "conformed/supermetrics/adls_ga_load_times_fr_test.csv"


def test_get_promoted_adls_path_dir_starts_with_slash():
    adls_path_dir_starts_with_slash = "/raw/supermetrics/adls_ga_load_times_fr_test/"
    flow = ADLSToAzureSQL(name="test", adls_path=adls_path_dir_starts_with_slash)
    promoted_path = flow.get_promoted_path(env="conformed")
    assert promoted_path == "conformed/supermetrics/adls_ga_load_times_fr_test.csv"


def test_df_to_csv_task():
    d = {"col1": ["rat", "\tdog"], "col2": ["cat", 4]}
    df = pd.DataFrame(data=d)
    assert df["col1"].astype(str).str.contains("\t")[1] == True
    task = df_to_csv_task
    task.run(df, path="result.csv", remove_tab=True)
    assert df["col1"].astype(str).str.contains("\t")[1] != True


def test_df_to_csv_task_none(caplog):
    df = None
    task = df_to_csv_task
    path = "result_none.csv"
    task.run(df, path=path, remove_tab=False)
    assert "DataFrame is None" in caplog.text
    assert os.path.isfile(path) == False


@pytest.mark.dtypes
def test_check_dtypes_sort():
    d = {"col1": ["rat", "cat"], "col2": [3, 4]}
    df = pd.DataFrame(data=d)
    dtypes = {
        "col1": "varchar(6)",
        "col2": "varchar(6)",
    }
    task = check_dtypes_sort

    n_dtypes = task.run(df=df, dtypes=dtypes)
    assert list(dtypes.keys()) == list(n_dtypes.keys())

    dtypes = {
        "col2": "varchar(6)",
        "col1": "varchar(6)",
    }
    n_dtypes = task.run(df=df, dtypes=dtypes)
    assert list(dtypes.keys()) != list(n_dtypes.keys())

    dtypes = {
        "col1": "varchar(6)",
        "col3": "varchar(6)",
    }
    try:
        n_dtypes = task.run(df=df, dtypes=dtypes)
        assert False
    except signals.FAIL:
        assert True



import json

import pandas as pd
import pytest

from viadot.sources import Hubspot
from viadot.task_utils import credentials_loader

CREDENTIALS = credentials_loader.run(credentials_secret="HUBSPOT")
HUBSPOT = Hubspot(credentials=CREDENTIALS)


@pytest.fixture(scope="session")
def var_dictionary():
    variables = {
        "endpoint": "contacts",
        "filters": [
            {
                "filters": [
                    {
                        "propertyName": "createdate",
                        "operator": "BETWEEN",
                        "highValue": "1642636800000",
                        "value": "1642535000000",
                    },
                    {
                        "propertyName": "email",
                        "operator": "CONTAINS_TOKEN",
                        "value": "*@seznam.cz",
                    },
                ]
            }
        ],
        "properties": [
            "recent_conversion_date",
            "recent_conversion_event_name",
            "hs_analytics_num_visits",
        ],
    }
    yield variables


def test_clean_special_characters():
    test_value = "762##28cd7-e$69d-4708-be31-726bb!859befd"
    clean_chars = HUBSPOT.clean_special_characters(value=test_value)
    assert clean_chars == "762%2523%252328cd7-e%252469d-4708-be31-726bb%2521859befd"


def test_get_api_url(var_dictionary):
    api_url = HUBSPOT.get_api_url(
        endpoint=var_dictionary["endpoint"],
        filters=var_dictionary["filters"],
        properties=var_dictionary["properties"],
    )
    assert (
        api_url
        == "https://api.hubapi.com/crm/v3/objects/contacts/search/?limit=100&properties=recent_conversion_date,recent_conversion_event_name,hs_analytics_num_visits&"
    )


def test_get_api_body(var_dictionary):
    api_body = HUBSPOT.get_api_body(filters=var_dictionary["filters"])

    assert (
        api_body
        == '{"filterGroups": [{"filters": [{"propertyName": "createdate", "operator": "BETWEEN", "highValue": "1642636800000", "value": "1642535000000"}, {"propertyName": "email", "operator": "CONTAINS_TOKEN", "value": "*@seznam.cz"}]}], "limit": 100}'
    )


def test_to_json(var_dictionary):
    api_url = HUBSPOT.get_api_url(
        endpoint=var_dictionary["endpoint"],
        filters=var_dictionary["filters"],
        properties=var_dictionary["properties"],
    )

    api_body = HUBSPOT.get_api_body(filters=var_dictionary["filters"])

    trigger = HUBSPOT.to_json(url=api_url, body=api_body, method="POST")

    assert isinstance(trigger, dict)



import pandas
import pytest

from viadot.sources.sqlite import SQLite

TABLE = "test"


@pytest.fixture(scope="session")
def sqlite():
    sqlite = SQLite(credentials=dict(db_name="testfile.sqlite"), query_timeout=5)

    yield sqlite

    sqlite.run(f"DROP TABLE {TABLE}")


def test_conn_str(sqlite):
    DRIVER = "/usr/lib/x86_64-linux-gnu/odbc/libsqlite3odbc.so"
    assert (
        sqlite.conn_str
        == f"DRIVER={{{DRIVER}}};SERVER=localhost;DATABASE=testfile.sqlite;"
    )


def test_create_table(sqlite):
    dtypes = {"country": "VARCHAR(100)", "sales": "FLOAT(24)"}
    result = sqlite.create_table(table=TABLE, dtypes=dtypes, if_exists="replace")
    assert result is True


def test_insert_into_sql(sqlite, DF):
    sql = sqlite.insert_into(TABLE, DF)

    assert "('italy', 100)" in sql
    assert sql[-1] == ";"

    results = sqlite.run(f"SELECT * FROM {TABLE}")
    df = pandas.DataFrame.from_records(results, columns=["country", "sales"])
    assert df["sales"].sum() == 230


def test_check_if_table_exists(sqlite):
    exists = sqlite._check_if_table_exists(TABLE)
    assert exists == True
    not_exists = sqlite._check_if_table_exists("test_table")
    assert not_exists == False



import os
from unittest import mock

import pandas as pd
import pendulum
import pytest
from prefect.tasks.secrets import PrefectSecret

from viadot.flows import SharepointToADLS
from viadot.tasks import AzureDataLakeRemove

ADLS_FILE_NAME = str(pendulum.now("utc")) + ".csv"
ADLS_DIR_PATH = "raw/tests/"
CREDENTIALS_SECRET = PrefectSecret("AZURE_DEFAULT_ADLS_SERVICE_PRINCIPAL_SECRET").run()
DATA = {"country": [1, 2], "sales": [3, 4]}


@mock.patch(
    "viadot.tasks.SharepointToDF.run",
    return_value=pd.DataFrame(data=DATA),
)
@pytest.mark.run
def test_sharepoint_to_adls_run_flow(mocked_class):
    flow = SharepointToADLS(
        "test_sharepoint_to_adls_run_flow",
        output_file_extension=".csv",
        adls_sp_credentials_secret=CREDENTIALS_SECRET,
        adls_dir_path=ADLS_DIR_PATH,
        adls_file_name=ADLS_FILE_NAME,
    )
    result = flow.run()
    assert result.is_successful()
    os.remove("test_sharepoint_to_adls_run_flow.csv")
    os.remove("test_sharepoint_to_adls_run_flow.json")


@mock.patch(
    "viadot.tasks.SharepointToDF.run",
    return_value=pd.DataFrame(data=DATA),
)
@pytest.mark.run
def test_sharepoint_to_adls_run_flow_overwrite_true(mocked_class):
    flow = SharepointToADLS(
        "test_sharepoint_to_adls_run_flow_overwrite_true",
        output_file_extension=".csv",
        adls_sp_credentials_secret=CREDENTIALS_SECRET,
        adls_dir_path=ADLS_DIR_PATH,
        adls_file_name=ADLS_FILE_NAME,
        overwrite_adls=True,
    )
    result = flow.run()
    assert result.is_successful()
    os.remove("test_sharepoint_to_adls_run_flow_overwrite_true.csv")
    os.remove("test_sharepoint_to_adls_run_flow_overwrite_true.json")


@mock.patch(
    "viadot.tasks.SharepointToDF.run",
    return_value=pd.DataFrame(data=DATA),
)
@pytest.mark.run
def test_sharepoint_to_adls_run_flow_overwrite_false(mocked_class):
    flow = SharepointToADLS(
        "test_sharepoint_to_adls_run_flow_overwrite_false",
        output_file_extension=".csv",
        adls_sp_credentials_secret=CREDENTIALS_SECRET,
        adls_dir_path=ADLS_DIR_PATH,
        adls_file_name=ADLS_FILE_NAME,
        overwrite_adls=False,
    )
    result = flow.run()

    assert result.is_failed()
    os.remove("test_sharepoint_to_adls_run_flow_overwrite_false.csv")
    os.remove("test_sharepoint_to_adls_run_flow_overwrite_false.json")



import os
from datetime import datetime
from pathlib import Path
from typing import Any, Dict, List, Literal

import pendulum
from prefect import Flow
from prefect.backend import set_key_value
from prefect.utilities import logging

from viadot.task_utils import (
    add_ingestion_metadata_task,
    anonymize_df,
    df_get_data_types_task,
    df_map_mixed_dtypes_for_parquet,
    df_to_csv,
    df_to_parquet,
    dtypes_to_json_task,
    update_dtypes_dict,
)
from viadot.tasks import AzureDataLakeUpload, CustomerGaugeToDF

logger = logging.get_logger(__name__)


class CustomerGaugeToADLS(Flow):
    def __init__(
        self,
        name: str,
        endpoint: Literal["responses", "non-responses"] = None,
        endpoint_url: str = None,
        total_load: bool = True,
        cursor: int = None,
        pagesize: int = 1000,
        date_field: Literal[
            "date_creation", "date_order", "date_sent", "date_survey_response"
        ] = None,
        start_date: datetime = None,
        end_date: datetime = None,
        customer_gauge_credentials_secret: str = "CUSTOMER-GAUGE",
        anonymize: bool = False,
        columns_to_anonymize: List[str] = None,
        anonymize_method: Literal["mask", "hash"] = "mask",
        anonymize_value: str = "***",
        date_column: str = None,
        days: int = None,
        output_file_extension: str = ".parquet",
        adls_dir_path: str = None,
        local_file_path: str = None,
        adls_file_name: str = None,
        vault_name: str = None,
        adls_sp_credentials_secret: str = None,
        overwrite_adls: bool = False,
        if_exists: str = "replace",
        timeout: int = 3600,
        *args: List[Any],
        **kwargs: Dict[str, Any]
    ):
        """
        Flow for downloading data from the Customer Gauge's endpoints (Responses and Non-Responses) via API to a CSV or Parquet file.
        The data anonimization is optional.Then upload it to Azure Data Lake.

        Args:
            name (str): The name of the flow.
            endpoint (Literal["responses", "non-responses"], optional): Indicate which endpoint to connect. Defaults to None.
            endpoint_url (str, optional): Full URL for pointing to specific endpoint. Defaults to None.
            total_load (bool, optional): Indicate whether to download the data to the latest. If 'False', only one API call is executed (up to 1000 records).
                Defaults to True.
            cursor (int, optional): Cursor value to navigate to the page. Defaults to None.
            pagesize (int, optional): Number of responses (records) returned per page, max value = 1000. Defaults to 1000.
            date_field (Literal["date_creation", "date_order", "date_sent", "date_survey_response"], optional): Specifies the date type which filter date range.
                Defaults to None.
            start_date (datetime, optional): Defines the period start date in yyyy-mm-dd format. Defaults to None.
            end_date (datetime, optional): Defines the period end date in yyyy-mm-dd format. Defaults to None.
            customer_gauge_credentials_secret (str, optional): The name of the Azure Key Vault secret containing a dictionary with ['client_id', 'client_secret'].
                Defaults to "CUSTOMER-GAUGE".
            vault_name (str, optional): The name of the vault from which to obtain the secret. Defaults to None.
            anonymize (bool, optional): Indicates if anonymize selected columns. Defaults to False.
            columns_to_anonymize (List[str], optional): List of columns to anonymize. Defaults to None.
            anonymize_method  (Literal["mask", "hash"], optional): Method of anonymizing data. "mask" -> replace the data with "value" arg.
                "hash" -> replace the data with the hash value of an object (using `hash()` method). Defaults to "mask".
            anonymize_value (str, optional): Value to replace the data. Defaults to "***".
            date_column (str, optional): Name of the date column used to identify rows that are older than a specified number of days. Defaults to None.
            days (int, optional): The number of days beyond which we want to anonymize the data, e.g. older that 2 years can be: 2*365. Defaults to None.
            output_file_extension (str, optional): Output file extension - to allow selection of .csv for data
                which is not easy to handle with parquet. Defaults to ".parquet".
            adls_dir_path (str, optional): Azure Data Lake destination folder/catalog path. Defaults to None.
            local_file_path (str, optional): Local destination path. Defaults to None.
            adls_file_name (str, optional): Name of file in ADLS. Defaults to None.
            adls_sp_credentials_secret (str, optional): The name of the Azure Key Vault secret containing a dictionary with
                ACCOUNT_NAME and Service Principal credentials (TENANT_ID, CLIENT_ID, CLIENT_SECRET) for the Azure Data Lake.
                Defaults to None.
            overwrite_adls (bool, optional): Whether to overwrite files in the lake. Defaults to False.
            if_exists (str, optional): What to do if the file exists. Defaults to "replace".
            timeout (int, optional): The time (in seconds) to wait while running this task before a timeout occurs. Defaults to 3600.
        """
        # CustomerGaugeToDF
        self.endpoint = endpoint
        self.endpoint_url = endpoint_url
        self.total_load = total_load
        self.cursor = cursor
        self.pagesize = pagesize
        self.date_field = date_field
        self.start_date = start_date
        self.end_date = end_date
        self.customer_gauge_credentials_secret = customer_gauge_credentials_secret

        # anonymize_df
        self.anonymize = anonymize
        self.columns_to_anonymize = columns_to_anonymize
        self.anonymize_method = anonymize_method
        self.anonymize_value = anonymize_value
        self.date_column = date_column
        self.days = days

        # AzureDataLakeUpload
        self.adls_file_name = adls_file_name
        self.adls_dir_path = adls_dir_path
        self.local_file_path = local_file_path
        self.overwrite = overwrite_adls
        self.vault_name = vault_name
        self.adls_sp_credentials_secret = adls_sp_credentials_secret
        self.if_exists = if_exists
        self.output_file_extension = output_file_extension
        self.timeout = timeout
        self.now = str(pendulum.now("utc"))

        self.local_file_path = (
            local_file_path or self.slugify(name) + self.output_file_extension
        )
        self.local_json_path = self.slugify(name) + ".json"
        self.adls_dir_path = adls_dir_path

        if adls_file_name is not None:
            self.adls_file_path = os.path.join(adls_dir_path, adls_file_name)
            self.adls_schema_file_dir_file = os.path.join(
                adls_dir_path, "schema", Path(adls_file_name).stem + ".json"
            )
        else:
            self.adls_file_path = os.path.join(
                adls_dir_path, self.now + self.output_file_extension
            )
            self.adls_schema_file_dir_file = os.path.join(
                adls_dir_path, "schema", self.now + ".json"
            )

        super().__init__(*args, name=name, **kwargs)

        self.gen_flow()

    @staticmethod
    def slugify(name):
        return name.replace(" ", "_").lower()

    def gen_flow(self) -> Flow:
        customer_gauge_df_task = CustomerGaugeToDF(
            timeout=self.timeout, endpoint=self.endpoint, endpoint_url=self.endpoint_url
        )

        customerg_df = customer_gauge_df_task.bind(
            total_load=self.total_load,
            cursor=self.cursor,
            pagesize=self.pagesize,
            date_field=self.date_field,
            start_date=self.start_date,
            end_date=self.end_date,
            vault_name=self.vault_name,
            credentials_secret=self.customer_gauge_credentials_secret,
            flow=self,
        )

        if self.anonymize == True:
            anonymized_df = anonymize_df.bind(
                customerg_df,
                columns=self.columns_to_anonymize,
                method=self.anonymize_method,
                value=self.anonymize_value,
                date_column=self.date_column,
                days=self.days,
                flow=self,
            )

            df_with_metadata = add_ingestion_metadata_task.bind(
                anonymized_df, flow=self
            )
        else:
            df_with_metadata = add_ingestion_metadata_task.bind(customerg_df, flow=self)

        dtypes_dict = df_get_data_types_task.bind(df_with_metadata, flow=self)
        df_mapped = df_map_mixed_dtypes_for_parquet.bind(
            df_with_metadata, dtypes_dict, flow=self
        )

        if self.output_file_extension == ".parquet":
            df_to_file = df_to_parquet.bind(
                df=df_mapped,
                path=self.local_file_path,
                if_exists=self.if_exists,
                flow=self,
            )
        else:
            df_to_file = df_to_csv.bind(
                df=df_with_metadata,
                path=self.local_file_path,
                if_exists=self.if_exists,
                flow=self,
            )

        file_to_adls_task = AzureDataLakeUpload(timeout=self.timeout)
        file_to_adls_task.bind(
            from_path=self.local_file_path,
            to_path=self.adls_file_path,
            overwrite=self.overwrite,
            sp_credentials_secret=self.adls_sp_credentials_secret,
            flow=self,
        )

        dtypes_updated = update_dtypes_dict(dtypes_dict, flow=self)
        dtypes_to_json_task.bind(
            dtypes_dict=dtypes_updated, local_json_path=self.local_json_path, flow=self
        )
        json_to_adls_task = AzureDataLakeUpload(timeout=self.timeout)
        json_to_adls_task.bind(
            from_path=self.local_json_path,
            to_path=self.adls_schema_file_dir_file,
            overwrite=self.overwrite,
            sp_credentials_secret=self.adls_sp_credentials_secret,
            flow=self,
        )

        file_to_adls_task.set_upstream(df_to_file, flow=self)
        json_to_adls_task.set_upstream(dtypes_to_json_task, flow=self)
        set_key_value(key=self.adls_dir_path, value=self.adls_file_path)



from typing import Any, Dict, List, Literal

from prefect import Flow

from viadot.task_utils import df_to_csv
from viadot.tasks import AzureDataLakeUpload
from viadot.tasks.mysql_to_df import MySqlToDf


class MySqlToADLS(Flow):
    def __init__(
        self,
        name: str,
        country_short: Literal["AT", "DE", "CH", None],
        query: str = None,
        sqldb_credentials_secret: str = None,
        vault_name: str = None,
        file_path: str = None,
        sep: str = "\t",
        to_path: str = None,
        if_exists: Literal["replace", "append", "delete"] = "replace",
        overwrite_adls: bool = True,
        sp_credentials_secret: str = None,
        credentials_secret: str = None,
        timeout: int = 3600,
        *args: List[any],
        **kwargs: Dict[str, Any]
    ):
        """
        Flow for downloading data from MySQL to csv file, then uploading it to ADLS.

        Args:
            name (str): The name of the flow.
            country_short (str): Country short to extract proper credentials.
            query (str): Query to perform on a database. Defaults to None.
            sqldb_credentials_secret (str, optional): The name of the Azure Key Vault secret containing a dictionary with
            MySQL Database credentials. Defaults to None.
            vault_name (str, optional): The name of the vault from which to obtain the secrets. Defaults to None.
            file_path (str, optional): Local destination path. Defaults to None.
            sep (str, optional): The delimiter for the output CSV file. Defaults to "\t".
            to_path (str): The path to an ADLS file. Defaults to None.
            if_exists (Literal, optional): What to do if the table exists. Defaults to "replace".
            overwrite_adls  (str, optional): Whether to overwrite_adls  the destination file. Defaults to True.
            sp_credentials_secret (str, optional): The name of the Azure Key Vault secret containing a dictionary with
            ACCOUNT_NAME and Service Principal credentials (TENANT_ID, CLIENT_ID, CLIENT_SECRET). Defaults to None.
            credentials_secret (str, optional): Key Vault name. Defaults to None.
            columns_to_clean (List(str), optional): Select columns to clean, used with remove_special_characters.
            If None whole data frame will be processed. Defaults to None.
            timeout(int, optional): The amount of time (in seconds) to wait while running this task before
                a timeout occurs. Defaults to 3600.
        """

        # Connect to sql
        self.country_short = country_short
        self.query = query
        self.sqldb_credentials_secret = sqldb_credentials_secret
        self.vault_name = vault_name
        self.overwrite_adls = overwrite_adls

        # Upload to ADLS
        self.file_path = file_path
        self.sep = sep
        self.to_path = to_path
        self.if_exists = if_exists
        self.sp_credentials_secret = sp_credentials_secret
        self.credentials_secret = credentials_secret
        self.timeout = timeout

        super().__init__(*args, name=name, **kwargs)

        self.gen_flow()

    def gen_flow(self) -> Flow:
        df_task = MySqlToDf(country_short=self.country_short, timeout=self.timeout)
        df = df_task.bind(
            credentials_secret=self.credentials_secret, query=self.query, flow=self
        )

        create_csv = df_to_csv.bind(
            df,
            path=self.file_path,
            sep=self.sep,
            if_exists=self.if_exists,
            flow=self,
        )

        file_to_adls_task = AzureDataLakeUpload(timeout=self.timeout)
        adls_upload = file_to_adls_task.bind(
            from_path=self.file_path,
            to_path=self.to_path,
            overwrite=self.overwrite_adls,
            sp_credentials_secret=self.sp_credentials_secret,
            flow=self,
        )

        create_csv.set_upstream(df, flow=self)
        adls_upload.set_upstream(create_csv, flow=self)



import pandas as pd

from viadot.utils import APIError, handle_api_response

from .base import Source


class Eurostat(Source):
    """
    Class for creating instance of Eurostat connector to REST API by HTTPS response (no credentials required).
    """

    def __init__(
        self,
        dataset_code: str,
        params: dict = None,
        base_url: str = "https://ec.europa.eu/eurostat/api/dissemination/statistics/1.0/data/",
        *args,
        **kwargs,
    ):
        """
        It is using HTTPS REST request to pull the data.
        No API registration or API key are required.
        Data will pull based on parameters provided in dynamic part of the url.

        Example of url: https://ec.europa.eu/eurostat/api/dissemination/statistics/1.0/data/TEIBS020/?format=JSON&lang=EN&indic=BS-CSMCI-BAL
        Static part: https://ec.europa.eu/eurostat/api/dissemination/statistics/1.0/data
        Dynamic part: /TEIBS020/?format=JSON&lang=EN&indic=BS-CSMCI-BAL

        Please note that for one dataset there are usually multiple data regarding different subjects.
        In order to retrive data that you are interested in you have to provide parameters with codes into 'params'.

        Args:
            dataset_code (str): The code of eurostat dataset that we would like to upload.
            params (Dict[str], optional):
                A dictionary with optional URL parameters. The key represents the parameter id, while the value is the code
                for a specific parameter, for example: params = {'unit': 'EUR'} where "unit" is the parameter that you would like to set
                and "EUR" is the code of the specific parameter. You can add more than one parameter, but only one code per parameter!
                So you CAN NOT provide list of codes as in example 'params = {'unit': ['EUR', 'USD', 'PLN']}'
                These parameters are REQUIRED in most cases to pull a specific dataset from the API.
                Both parameter and code has to be provided as a string! Defaults to None.
            base_url (str): The base URL used to access the Eurostat API. This parameter specifies the root URL for all requests made to the API.
                It should not be modified unless the API changes its URL scheme.
                Defaults to "https://ec.europa.eu/eurostat/api/dissemination/statistics/1.0/data/"
        Raises:
            TypeError: If self.params is different type than a dictionary.
        """

        self.dataset_code = dataset_code
        self.params = params
        if not isinstance(self.params, dict) and self.params is not None:
            raise TypeError("Params should be a dictionary.")
        self.base_url = f"{base_url}{self.dataset_code}?format=JSON&lang=EN"

        super().__init__(*args, **kwargs)

    def get_parameters_codes(self) -> dict:
        """Function for getting available parameters with codes from dataset.

        Raises:
            ValueError: If the response from the API is empty or invalid.

        Returns:
            Dict: Key is parameter and value is a list of available codes for specific parameter.
        """

        try:
            response = handle_api_response(self.base_url)
            data = response.json()
        except APIError:
            self.logger.error(
                f"Failed to fetch data for {self.dataset_code}, please check correctness of dataset code!"
            )
            raise ValueError("DataFrame is empty!")

        # getting list of available parameters
        available_params = data["id"]

        # dictionary from JSON with keys and reletad codes values
        dimension = data["dimension"]

        # Assigning list of available codes to specific parameters
        params_and_codes = {}
        for key in available_params:
            if key in dimension:
                codes = list(dimension[key]["category"]["index"].keys())
                params_and_codes[key] = codes
        return params_and_codes

    def make_params_validation(self):
        """Function for validation of given parameters in comparison
        to parameteres and their codes from JSON.

        Raises:
            ValueError: If any of the self.params keys or values is not a string or
            any of them is not available for specific dataset.
        """

        # In order to make params validation, first we need to get params_and_codes.
        key_codes = self.get_parameters_codes()

        # Validation of type of values
        for key, val in self.params.items():
            if not isinstance(key, str) or not isinstance(val, str):
                self.logger.error(
                    "You can provide only one code per one parameter as 'str' in params!\n"
                    "CORRECT: params = {'unit': 'EUR'} | INCORRECT: params = {'unit': ['EUR', 'USD', 'PLN']}"
                )
                raise ValueError("Wrong structure of params!")

        if key_codes is not None:
            # Conversion keys and values on lowwer cases by using casefold
            key_codes_after_conversion = {
                k.casefold(): [v_elem.casefold() for v_elem in v]
                for k, v in key_codes.items()
            }
            params_after_conversion = {
                k.casefold(): v.casefold() for k, v in self.params.items()
            }

            # comparing keys and values
            non_available_keys = [
                key
                for key in params_after_conversion.keys()
                if key not in key_codes_after_conversion
            ]
            non_available_codes = [
                value
                for key, value in params_after_conversion.items()
                if key in key_codes_after_conversion.keys()
                and value not in key_codes_after_conversion[key]
            ]

            # error loggers
            if non_available_keys:
                self.logger.error(
                    f"Parameters: '{' | '.join(non_available_keys)}' are not in dataset. Please check your spelling!\n"
                    f"Possible parameters: {' | '.join(key_codes.keys())}"
                )
            if non_available_codes:
                self.logger.error(
                    f"Parameters codes: '{' | '.join(non_available_codes)}' are not available. Please check your spelling!\n"
                    f"You can find everything via link: https://ec.europa.eu/eurostat/databrowser/view/{self.dataset_code}/default/table?lang=en"
                )
            raise ValueError("DataFrame is empty!")

    def eurostat_dictionary_to_df(self, *signals: list) -> pd.DataFrame:
        """Function for creating DataFrame from JSON pulled from Eurostat.

        Returns:
            pd.DataFrame: With 4 columns: index, geo, time, indicator.
        """

        class T_SIGNAL:
            signal_keys_list: list
            signal_index_list: list
            signal_label_list: list
            signal_name: str

        # Dataframe creation
        columns0 = signals[0].copy()
        columns0.append("indicator")
        df = pd.DataFrame(columns=columns0)
        indicator_list = []
        index_list = []
        signal_lists = []

        # get the dictionary from the inputs
        eurostat_dictionary = signals[-1]

        for signal in signals[0]:
            signal_struct = T_SIGNAL()
            signal_struct.signal_name = signal
            signal_struct.signal_keys_list = list(
                eurostat_dictionary["dimension"][signal]["category"]["index"].keys()
            )
            signal_struct.signal_index_list = list(
                eurostat_dictionary["dimension"][signal]["category"]["index"].values()
            )
            signal_label_dict = eurostat_dictionary["dimension"][signal]["category"][
                "label"
            ]
            signal_struct.signal_label_list = [
                signal_label_dict[i] for i in signal_struct.signal_keys_list
            ]
            signal_lists.append(signal_struct)

        col_signal_temp = []
        row_signal_temp = []
        for row_index, row_label in zip(
            signal_lists[0].signal_index_list, signal_lists[0].signal_label_list
        ):  # rows
            for col_index, col_label in zip(
                signal_lists[1].signal_index_list, signal_lists[1].signal_label_list
            ):  # cols
                index = str(
                    col_index + row_index * len(signal_lists[1].signal_label_list)
                )
                if index in eurostat_dictionary["value"].keys():
                    index_list.append(index)
                    col_signal_temp.append(col_label)
                    row_signal_temp.append(row_label)

        indicator_list = [eurostat_dictionary["value"][i] for i in index_list]
        df.indicator = indicator_list
        df[signal_lists[1].signal_name] = col_signal_temp
        df[signal_lists[0].signal_name] = row_signal_temp

        return df

    def get_data_frame_from_response(self) -> pd.DataFrame:
        """Function responsible for getting response, creating DataFrame using method 'eurostat_dictionary_to_df'
           with validation of provided parameters and their codes if needed.

        Raises:
            APIError: If there is an error with the API request.
            ValueError: If the resulting DataFrame is empty.

        Returns:
            pd.DataFrame: Final DataFrame or raise prefect.logger.error, if issues occur.
        """

        try:
            response = handle_api_response(self.base_url, params=self.params)
            data = response.json()
            data_frame = self.eurostat_dictionary_to_df(["geo", "time"], data)

            if data_frame.empty:
                raise ValueError
        except (APIError, ValueError):
            self.make_params_validation()

        # merging data_frame with label and last updated date
        label_col = pd.Series(str(data["label"]), index=data_frame.index, name="label")
        last_updated__col = pd.Series(
            str(data["updated"]),
            index=data_frame.index,
            name="updated",
        )
        data_frame = pd.concat([data_frame, label_col, last_updated__col], axis=1)
        return data_frame



import os
from pathlib import Path
import shutil
from typing import Dict, List, Union

from prefect import Flow, task
from prefect.tasks.shell import ShellTask
from prefect.triggers import any_successful

from viadot.tasks import CloneRepo, AzureKeyVaultSecret, LumaIngest


@task(trigger=any_successful)
def _cleanup_repo(dbt_repo_dir_name: str) -> None:
    """
    Remove a repo folder.

    Args:
        dbt_repo_dir_name (str): The name of the temporary folder.
    """
    shutil.rmtree(dbt_repo_dir_name, ignore_errors=True)  # Delete folder on run


@task(trigger=any_successful)
def custom_shell_task(name: str, command: str, helper_script: str = None) -> None:
    """
    Task created to run ShellTask and apply `trigger` on it. For regular ShellTask it is not possible.

    Args:
        name (str): The name of the flow.
        command (str): Shell command to run.
        helper_script (str, optional): Path to local path repo. Defaults to None.
    """
    ShellTask(
        name=name,
        command=command,
        helper_script=helper_script,
        return_all=True,
        stream_output=True,
    ).run()


class TransformAndCatalogToLuma(Flow):
    """Build specified dbt model(s) and upload the generated metadata to Luma catalog."""

    def __init__(
        self,
        name: str,
        dbt_project_path: str = "tmp_dbt_repo_dir",
        dbt_repo_url: str = None,
        dbt_repo_url_secret: str = None,
        dbt_repo_branch: str = None,
        token: str = None,
        token_secret: str = None,
        local_dbt_repo_path: str = None,
        dbt_selects: Dict[str, str] = None,
        dbt_target: str = None,
        metadata_dir_path: Union[str, Path] = None,
        luma_url: str = "http://localhost",
        luma_url_secret: str = None,
        vault_name: str = None,
        *args,
        **kwargs,
    ) -> List[str]:
        """
        Build specified dbt model(s) and upload the generated metadata to Luma catalog.

        Args:
            name (str): The name of the Flow.
            dbt_project_path (str, optional): The path to the dbt project (the directory containing
                the `dbt_project.yml` file). Defaults to 'tmp_dbt_repo_dir'.
            dbt_repo_url (str, optional): The URL for cloning the dbt repo with relevant dbt project. Defaults to None.
            dbt_repo_url_secret (str, optional): Alternatively to above, the secret containing `dbt_repo_url`.
                Defaults to None.
            dbt_repo_branch (str, optional): The branch of the dbt repo to use. Defaults to None.
            token (str, optional): The personal access token. Defaults to None.
            token_secret (str, optional): The name of the secret storing the token. Defaults to None.
            local_dbt_repo_path (str, optional): The path where to clone the repo to. Defaults to None.
            dbt_selects (dict, optional): Valid [dbt node selection](https://docs.getdbt.com/reference/node-selection/syntax)
                expressions. Valid keys are `run`, `test`, and `source_freshness`. The test select expression is taken
                from run's, as long as run select is provided. Defaults to None.
            dbt_target (str): The dbt target to use. If not specified, the default dbt target (as specified in `profiles.yaml`)
                will be used. Defaults to None.
            metadata_dir_path (Union[str, Path]): The path to the directory containing metadata files.
                In the case of dbt, it's dbt project's `target` directory, which contains dbt artifacts
                (`sources.json`, `catalog.json`, `manifest.json`, and `run_results.json`). Defaults to None.
            luma_url (str, optional): The url of the Luma ingestion API. Defaults to "http://localhost".
            luma_url_secret (str, optional): The name of the secret storing the luma_url. Defaults to None.
            vault_name (str, optional): The name of the vault from which to obtain the secrets. Defaults to None.

        Returns:
            List[str]: Lines from stdout of the `upload_metadata` task as a list.

        Examples:
            # Build a single model
            ```python
            import os
            from viadot.flows import TransformAndCatalogToLuma

            my_dbt_project_path = os.path.expanduser("~/dbt/my_dbt_project")

            flow = TransformAndCatalogToLuma(
                name="Transform and Catalog to Luma",
                dbt_project_path=my_dbt_project_path,
                dbt_repo_url=my_dbt_repo_url,
                token=my_token,
                dbt_selects={
                    "run": "my_model",
                    "source_freshness": "source:schema.table",
                    "test": "my_model"},
                metadata_dir_path=f"{my_dbt_project_path}/target",
                luma_url="http://localhost"
            )
            flow.run()
            ```

            Some common `dbt_select` patterns:
            - build a model and all its downstream dependencies: `dbt_select="my_model+"`
            - build all models in a directory: `dbt_select="models/my_project"`
            ```
        """
        # DBTTask
        self.dbt_project_path = dbt_project_path
        self.dbt_target = dbt_target
        self.dbt_selects = dbt_selects

        # CloneRepo
        self.dbt_repo_url = dbt_repo_url
        self.dbt_repo_url_secret = dbt_repo_url_secret
        self.dbt_repo_branch = dbt_repo_branch
        self.token = token
        self.token_secret = token_secret
        self.local_dbt_repo_path = local_dbt_repo_path

        # LumaIngest
        self.metadata_dir_path = metadata_dir_path
        self.luma_url = luma_url
        self.luma_url_secret = luma_url_secret
        self.vault_name = vault_name

        super().__init__(*args, name=name, **kwargs)
        self.gen_flow()

    @staticmethod
    def slugify(name):
        return name.replace(" ", "_").lower()

    def gen_flow(self) -> Flow:
        azure_secret_task = AzureKeyVaultSecret()
        dbt_repo_url = self.dbt_repo_url or azure_secret_task.bind(
            self.dbt_repo_url_secret
        )
        local_dbt_repo_path = (
            os.path.expandvars(self.local_dbt_repo_path)
            if self.local_dbt_repo_path is not None
            else f"{self.dbt_project_path}"
        )

        clone_repo = CloneRepo(url=dbt_repo_url)
        clone = clone_repo.bind(
            url=dbt_repo_url,
            checkout_branch=self.dbt_repo_branch,
            token=self.token,
            token_secret=self.token_secret,
            path=local_dbt_repo_path,
            flow=self,
        )

        # dbt CLI does not handle passing --target=None
        dbt_target_option = (
            f"-t {self.dbt_target}" if self.dbt_target is not None else ""
        )

        # Clean up artifacts from previous runs (`target/` dir and packages)
        dbt_clean_up = ShellTask(
            name="dbt_task_clean",
            command=f"dbt clean",
            helper_script=f"cd {local_dbt_repo_path}",
            return_all=True,
            stream_output=True,
        ).bind(flow=self)

        pull_dbt_deps = ShellTask(
            name="dbt_task_deps",
            command=f"dbt deps",
            helper_script=f"cd {local_dbt_repo_path}",
            return_all=True,
            stream_output=True,
        ).bind(flow=self)

        run_select = self.dbt_selects.get("run")
        run_select_safe = f"-s {run_select}" if run_select is not None else ""

        run = ShellTask(
            name="dbt_task_run",
            command=f"dbt run {run_select_safe} {dbt_target_option}",
            helper_script=f"cd {local_dbt_repo_path}",
            return_all=True,
            stream_output=True,
        ).bind(flow=self)

        test_select = self.dbt_selects.get("test", run_select)
        test_select_safe = f"-s {test_select}" if test_select is not None else ""

        test = ShellTask(
            name="dbt_task_test",
            command=f"dbt test {test_select_safe} {dbt_target_option}",
            helper_script=f"cd {local_dbt_repo_path}",
            return_all=True,
            stream_output=True,
        ).bind(flow=self)

        # Generate docs
        # Produces `catalog.json`, `run-results.json`, and `manifest.json`

        generate_catalog_json = custom_shell_task.bind(
            name="dbt_task_docs_generate",
            command=f"dbt docs generate {dbt_target_option} --no-compile",
            helper_script=f"cd {self.dbt_project_path}",
            flow=self,
        )

        # Upload build metadata to Luma
        path_expanded = os.path.expandvars(self.metadata_dir_path)
        metadata_dir_path = Path(path_expanded)

        upload_metadata_luma = LumaIngest(
            name="luma_ingest_task",
            metadata_dir_path=metadata_dir_path,
            url=self.luma_url,
            dbt_project_path=self.dbt_project_path,
            credentials_secret=self.luma_url_secret,
            vault_name=self.vault_name,
        ).bind(flow=self)

        _cleanup_repo.bind(local_dbt_repo_path, flow=self)

        dbt_clean_up.set_upstream(clone, flow=self)
        pull_dbt_deps.set_upstream(dbt_clean_up, flow=self)
        run.set_upstream(pull_dbt_deps, flow=self)
        test.set_upstream(run, flow=self)
        generate_catalog_json.set_upstream(test, flow=self)
        upload_metadata_luma.set_upstream(generate_catalog_json, flow=self)
        _cleanup_repo.set_upstream(upload_metadata_luma, flow=self)



import os
from pathlib import Path
from typing import Any, Dict, List, Literal

import pendulum
from prefect import Flow
from prefect.backend import set_key_value

from viadot.task_utils import (
    add_ingestion_metadata_task,
    df_get_data_types_task,
    df_map_mixed_dtypes_for_parquet,
    df_to_parquet,
    dtypes_to_json_task,
    update_dtypes_dict,
)
from viadot.tasks import AzureDataLakeUpload, HubspotToDF


class HubspotToADLS(Flow):
    def __init__(
        self,
        name: str,
        endpoint: str,
        hubspot_credentials: dict = None,
        hubspot_credentials_key: str = "HUBSPOT",
        properties: List[Any] = [],
        filters: Dict[str, Any] = {},
        nrows: int = 1000,
        output_file_extension: str = ".parquet",
        local_file_path: str = None,
        adls_file_name: str = None,
        adls_dir_path: str = None,
        if_exists: Literal["replace", "append", "delete"] = "replace",
        overwrite_adls: bool = True,
        vault_name: str = None,
        sp_credentials_secret: str = None,
        *args: List[any],
        **kwargs: Dict[str, Any],
    ):
        """
        Flow for downloading data from Hubspot to parquet file, then uploading it to ADLS.

        Args:
            name (str): The name of the flow.
            endpoint (str): Full Hubspot endpoint ID (name) or type of schema that will be passed to the url.
            hubspot_credentials (dict): Credentials to Hubspot API. Defaults to None.
            hubspot_credentials_key (str, optional): Credential key to dictionary where credentials are stored (e.g. in local config). Defaults to "HUBSPOT".
            properties (List, optional): List of properties/columns that will be passed to the url. Defaults to [].
            filters (Dict, optional): Filters for the Hubspot API body in JSON format. Defaults to {}.
                                        - propertyName: columns for filtering
                                        - operator: operator for filtering from the list [IN, NOT_HAS_PROPERTY, LT, EQ, GT, NOT_IN, GTE, CONTAINS_TOKEN, HAS_PROPERTY, LTE, NOT_CONTAINS_TOKEN, BETWEEN, NEQ]
                                        - highValue - max value
                                        - value - min value

                                    Example below:

                                        filters = [
                                                {
                                                "filters": [
                                                    {
                                                    "propertyName": "createdate",
                                                    "operator": "BETWEEN",
                                                    "highValue": "2023-03-27",
                                                    "value": "2023-03-23"
                                                    }
                                                ]
                                                }
                                            ]

            nrows (int, optional): Maximum number of rows to be pulled. Defaults to 1000.
            local_file_path (str, optional): Local destination path. Defaults to None.
            adls_file_name (str, optional): Name of file in ADLS. Defaults to None.
            adls_dir_path (str, optional): Azure Data Lake destination folder/catalog path. Defaults to None.
            output_file_extension (str, optional): Output file extension. Defaults to ".parquet".
            if_exists (Literal, optional): What to do if the table exists. Defaults to "replace".
            overwrite_adls (str, optional): Whether to overwrite the destination file in ADLS. Defaults to True.
            vault_name (str, optional): The name of the vault from which to obtain the secrets. Defaults to None.
            sp_credentials_secret (str, optional): The name of the Azure Key Vault secret containing a dictionary with ACCOUNT_NAME and Service Principal credentials (TENANT_ID, CLIENT_ID, CLIENT_SECRET). Defaults to None.
        """

        self.endpoint = endpoint
        self.properties = properties
        self.filters = filters
        self.nrows = nrows
        self.hubspot_credentials = hubspot_credentials
        self.hubspot_credentials_key = hubspot_credentials_key
        self.output_file_extension = output_file_extension

        self.local_file_path = (
            local_file_path or self.slugify(name) + self.output_file_extension
        )
        self.local_json_path = self.slugify(name) + ".json"
        self.now = str(pendulum.now("utc"))
        self.adls_dir_path = adls_dir_path

        if adls_file_name is not None:
            self.adls_file_path = os.path.join(adls_dir_path, adls_file_name)
            self.adls_schema_file_dir_file = os.path.join(
                adls_dir_path, "schema", Path(adls_file_name).stem + ".json"
            )

        else:
            self.adls_file_path = os.path.join(
                adls_dir_path, self.now + self.output_file_extension
            )
            self.adls_schema_file_dir_file = os.path.join(
                adls_dir_path, "schema", self.now + ".json"
            )

        self.if_exists = if_exists
        self.overwrite_adls = overwrite_adls
        self.vault_name = vault_name
        self.sp_credentials_secret = sp_credentials_secret

        super().__init__(*args, name=name, **kwargs)
        self.gen_flow()

    @staticmethod
    def slugify(name):
        return name.replace(" ", "_").lower()

    def gen_flow(self) -> Flow:
        hubspot_to_df_task = HubspotToDF(
            hubspot_credentials=self.hubspot_credentials,
            hubspot_credentials_key=self.hubspot_credentials_key,
        )

        df = hubspot_to_df_task.bind(
            endpoint=self.endpoint,
            properties=self.properties,
            filters=self.filters,
            nrows=self.nrows,
            flow=self,
        )

        df_viadot_downloaded = add_ingestion_metadata_task.bind(df=df, flow=self)
        dtypes_dict = df_get_data_types_task.bind(df_viadot_downloaded, flow=self)

        df_to_be_loaded = df_map_mixed_dtypes_for_parquet(
            df_viadot_downloaded, dtypes_dict, flow=self
        )

        df_to_parquet_task = df_to_parquet.bind(
            df=df_to_be_loaded,
            path=self.local_file_path,
            if_exists=self.if_exists,
            flow=self,
        )

        file_to_adls_task = AzureDataLakeUpload()
        adls_upload = file_to_adls_task.bind(
            from_path=self.local_file_path,
            to_path=self.adls_file_path,
            overwrite=self.overwrite_adls,
            sp_credentials_secret=self.sp_credentials_secret,
            flow=self,
        )

        dtypes_updated = update_dtypes_dict(dtypes_dict, flow=self)
        dtypes_to_json_task.bind(
            dtypes_dict=dtypes_updated, local_json_path=self.local_json_path, flow=self
        )

        json_to_adls_task = AzureDataLakeUpload()
        json_to_adls_task.bind(
            from_path=self.local_json_path,
            to_path=self.adls_schema_file_dir_file,
            overwrite=self.overwrite_adls,
            sp_credentials_secret=self.sp_credentials_secret,
            vault_name=self.vault_name,
            flow=self,
        )

        df_viadot_downloaded.set_upstream(df, flow=self)
        dtypes_dict.set_upstream(df_viadot_downloaded, flow=self)
        df_to_be_loaded.set_upstream(dtypes_dict, flow=self)
        adls_upload.set_upstream(df_to_parquet_task, flow=self)

        dtypes_to_json_task.set_upstream(dtypes_updated, flow=self)
        json_to_adls_task.set_upstream(dtypes_to_json_task, flow=self)

        set_key_value(key=self.adls_dir_path, value=self.adls_file_path)



import os
from csv import reader
from io import StringIO
from unittest import mock

import pandas as pd
from prefect.storage import Local

from viadot.config import local_config
from viadot.flows import SupermetricsToAzureSQL

CWD = os.getcwd()
adls_dir_path = "raw/supermetrics"
STORAGE = Local(path=CWD)
SCHEMA = "sandbox"
TABLE = "test_supermetrics"


def test_supermetrics_to_azure_sql_init():
    instance = SupermetricsToAzureSQL(
        "test_name",
        ds_id="example_id",
        ds_accounts="example_accounts",
        ds_user="example_user",
        fields=["filed", "field2"],
    )

    assert instance
    assert instance.__dict__["ds_id"] == "example_id"


def test_supermetrics_to_azure_sql_run_flow_mock():
    with mock.patch.object(
        SupermetricsToAzureSQL, "run", return_value=True
    ) as mock_method:
        credentials = local_config.get("SUPERMETRICS")

        flow = SupermetricsToAzureSQL(
            "test_supermetrics",
            ds_id="GA",
            ds_segments=[
                "R1fbzFNQQ3q_GYvdpRr42w",
                "I8lnFFvdSFKc50lP7mBKNA",
                "Lg7jR0VWS5OqGPARtGYKrw",
                "h8ViuGLfRX-cCL4XKk6yfQ",
                "-1",
            ],
            ds_accounts=["8326007", "58338899"],
            date_range_type="last_month",
            ds_user=credentials["USER"],
            fields=[
                {"id": "Date"},
                {"id": "segment", "split": "column"},
                {"id": "AvgPageLoadTime_calc"},
            ],
            dtypes={
                "date": "DATE",
                "segment": "VARCHAR(255)",
                "AvgPageLoadTime_calc": "VARCHAR(255)",
            },
            settings={"avoid_sampling": "true"},
            order_columns="alphabetic",
            max_columns=10,
            max_rows=1,
            schema=SCHEMA,
            table=TABLE,
            local_file_path="test_supermetrics.csv",
            blob_path="tests/test.csv",
            storage=STORAGE,
        )

        flow.run()
        mock_method.assert_called_with()



import os
from pathlib import Path
from typing import Any, Dict, List

import pendulum
from prefect import Flow
from prefect.backend import set_key_value
from prefect.utilities import logging

from ..task_utils import (
    add_ingestion_metadata_task,
    cast_df_to_str,
    df_get_data_types_task,
    df_map_mixed_dtypes_for_parquet,
    df_to_csv,
    df_to_parquet,
    dtypes_to_json_task,
    update_dtypes_dict,
)
from ..tasks import AzureDataLakeUpload, MediatoolToDF

file_to_adls_task = AzureDataLakeUpload()
json_to_adls_task = AzureDataLakeUpload()

logger = logging.get_logger(__name__)


class MediatoolToADLS(Flow):
    def __init__(
        self,
        name: str,
        organization_ids: List[str] = None,
        media_entries_columns: List[str] = None,
        mediatool_credentials: dict = None,
        mediatool_credentials_key: str = "MEDIATOOL",
        vault_name: str = None,
        output_file_extension: str = ".parquet",
        adls_dir_path: str = None,
        local_file_path: str = None,
        adls_file_name: str = None,
        adls_sp_credentials_secret: str = None,
        overwrite_adls: bool = False,
        if_exists: str = "replace",
        *args: List[Any],
        **kwargs: Dict[str, Any],
    ):
        """
        Flow for downloading data from the Mediatool platform via API to a CSV or Parquet file.
        Then upload it to Azure Data Lake.

        Args:
            name (str): The name of the flow.
            organization_ids (List[str], optional): List of organization IDs. Defaults to None.
            media_entries_columns (List[str], optional): Columns to get from media entries. Defaults to None.
            mediatool_credentials (dict, optional): Dictionary containing Mediatool credentials. Defaults to None.
            mediatool_credentials_key (str, optional): Credential key to dictionary where credentials are stored (e.g. in local config).
                Defaults to "MEDIATOOL".
            vault_name (str, optional): The name of the vault from which to obtain the secrets. Defaults to None.
            output_file_extension (str, optional): Output file extension - to allow selection of .csv for data
                which is not easy to handle with parquet. Defaults to ".parquet".
            adls_dir_path (str, optional): Azure Data Lake destination folder/catalog path. Defaults to None.
            local_file_path (str, optional): Local destination path. Defaults to None.
            adls_file_name (str, optional): Name of file in ADLS. Defaults to None.
            adls_sp_credentials_secret (str, optional): The name of the Azure Key Vault secret containing a dictionary with
                ACCOUNT_NAME and Service Principal credentials (TENANT_ID, CLIENT_ID, CLIENT_SECRET) for the Azure Data Lake.
                Defaults to None.
            overwrite_adls (bool, optional): Whether to overwrite files in the lake. Defaults to False.
            if_exists (str, optional): What to do if the file exists. Defaults to "replace".
        """
        # MediatoolToDF
        self.organization_ids = organization_ids
        self.mediatool_credentials = mediatool_credentials
        self.mediatool_credentials_key = mediatool_credentials_key
        self.media_entries_columns = media_entries_columns
        self.vault_name = vault_name

        # AzureDataLakeUpload
        self.overwrite = overwrite_adls
        self.adls_sp_credentials_secret = adls_sp_credentials_secret
        self.if_exists = if_exists
        self.output_file_extension = output_file_extension
        self.now = str(pendulum.now("utc"))

        self.local_file_path = (
            local_file_path or self.slugify(name) + self.output_file_extension
        )
        self.local_json_path = self.slugify(name) + ".json"
        self.adls_dir_path = adls_dir_path

        if adls_file_name is not None:
            self.adls_file_path = os.path.join(adls_dir_path, adls_file_name)
            self.adls_schema_file_dir_file = os.path.join(
                adls_dir_path, "schema", Path(adls_file_name).stem + ".json"
            )
        else:
            self.adls_file_path = os.path.join(
                adls_dir_path, self.now + self.output_file_extension
            )
            self.adls_schema_file_dir_file = os.path.join(
                adls_dir_path, "schema", self.now + ".json"
            )

        super().__init__(*args, name=name, **kwargs)

        self.gen_flow()

    @staticmethod
    def slugify(name):
        return name.replace(" ", "_").lower()

    def gen_flow(self) -> Flow:
        mediatool_to_df_task = MediatoolToDF(
            mediatool_credentials=self.mediatool_credentials,
            mediatool_credentials_key=self.mediatool_credentials_key,
        )

        df = mediatool_to_df_task.bind(
            organization_ids=self.organization_ids,
            media_entries_columns=self.media_entries_columns,
            flow=self,
        )

        df_with_metadata = add_ingestion_metadata_task.bind(df, flow=self)
        df_casted_to_str = cast_df_to_str(df_with_metadata, flow=self)
        dtypes_dict = df_get_data_types_task.bind(df_casted_to_str, flow=self)

        if self.output_file_extension == ".parquet":
            df_to_be_loaded = df_map_mixed_dtypes_for_parquet(
                df_casted_to_str, dtypes_dict, flow=self
            )
            df_to_file = df_to_parquet.bind(
                df=df_to_be_loaded,
                path=self.local_file_path,
                if_exists=self.if_exists,
                flow=self,
            )
        else:
            df_to_file = df_to_csv.bind(
                df=df_casted_to_str,
                path=self.local_file_path,
                if_exists=self.if_exists,
                flow=self,
            )

        file_to_adls_task.bind(
            from_path=self.local_file_path,
            to_path=self.adls_file_path,
            overwrite=self.overwrite,
            sp_credentials_secret=self.adls_sp_credentials_secret,
            flow=self,
        )

        dtypes_updated = update_dtypes_dict(dtypes_dict, flow=self)
        dtypes_to_json_task.bind(
            dtypes_dict=dtypes_updated, local_json_path=self.local_json_path, flow=self
        )

        json_to_adls_task.bind(
            from_path=self.local_json_path,
            to_path=self.adls_schema_file_dir_file,
            overwrite=self.overwrite,
            sp_credentials_secret=self.adls_sp_credentials_secret,
            flow=self,
        )

        file_to_adls_task.set_upstream(df_to_file, flow=self)
        json_to_adls_task.set_upstream(dtypes_to_json_task, flow=self)
        set_key_value(key=self.adls_dir_path, value=self.adls_file_path)



import json
import time
from typing import Any, Dict, List

import prefect
from prefect import Task
from prefect.tasks.secrets import PrefectSecret

from viadot.config import local_config
from viadot.sources.sftp import SftpConnector

from .azure_key_vault import AzureKeyVaultSecret


class SftpToDF(Task):
    def __init__(
        self,
        credentials: Dict[str, Any] = None,
        sftp_credentials_secret: str = None,
        vault_name: str = None,
        timeout: int = 3600,
        *args,
        **kwargs,
    ):
        """Task for downloading data from SFTP server to pandas DataFrame.

        Args:
            credentials (Dict[str, Any], optional): SFTP credentials. Defaults to None.
            sftp_credentials_secret (str, optional): The name of the Azure Key Vault secret containing a dictionary credentials for SFTP connection. Defaults to None.
            vault_name (str, optional): The name of the vault from which to obtain the secret. Defaults to None.
            timeout(int, optional): The amount of time (in seconds) to wait while running this task before
                a timeout occurs. Defaults to 3600.

        Returns: Pandas DataFrame
        """
        self.credentials = credentials
        self.sftp_credentials_secret = sftp_credentials_secret
        self.vault_name = vault_name

        super().__init__(
            name="SftpToDF",
            timeout=timeout,
            *args,
            **kwargs,
        )

    def __call__(self, *args, **kwargs):
        """Download file-like object from SFTP server to df"""
        return super().__call__(*args, **kwargs)

    def run(
        self,
        from_path: str = None,
        sep: str = "\t",
        columns: List[str] = None,
        **kwargs,
    ):
        logger = prefect.context.get("logger")
        if not self.credentials:
            if not self.sftp_credentials_secret:
                try:
                    self.sftp_credentials_secret = PrefectSecret("QWASI-SFTP").run()
                    logger.info("Loaded credentials from Key Vault.")
                except ValueError:
                    pass

            if self.sftp_credentials_secret:
                credentials_str = AzureKeyVaultSecret(
                    self.sftp_credentials_secret, vault_name=self.vault_name
                ).run()
                credentials_sftp = json.loads(credentials_str)
                logger.info("Loaded credentials from Key Vault.")
            else:
                credentials_sftp = local_config.get("QWASI-SFTP")
                logger.info("Loaded credentials from local source.")
        else:
            credentials_sftp = self.credentials

        sftp = SftpConnector(credentials_sftp=credentials_sftp)
        sftp.get_conn()
        logger.info("Connected to SFTP server.")
        time.sleep(1)
        df = sftp.to_df(file_name=from_path, sep=sep, columns=columns, **kwargs)
        logger.info("Succefully downloaded file from SFTP server.")
        return df


class SftpList(Task):
    def __init__(
        self,
        credentials: Dict[str, Any] = None,
        sftp_credentials_secret: str = None,
        vault_name: str = None,
        timeout: int = 3600,
        *args,
        **kwargs,
    ):
        """Task for listing files in SFTP server.

        Args:
            credentials (Dict[str, Any], optional): SFTP credentials. Defaults to None.
            sftp_credentials_secret (str, optional): The name of the Azure Key Vault secret containing a dictionary credentials for SFTP connection. Defaults to None.
            vault_name (str, optional): The name of the vault from which to obtain the secret. Defaults to None.
            timeout(int, optional): The amount of time (in seconds) to wait while running this task before
                a timeout occurs. Defaults to 3600.

        Returns:
            files_list (List): List of files in SFTP server.
        """
        self.credentials = credentials
        self.sftp_credentials_secret = sftp_credentials_secret
        self.vault_name = vault_name

        super().__init__(
            name="SftpList",
            timeout=timeout,
            *args,
            **kwargs,
        )

    def __call__(self, *args, **kwargs):
        """Task for listing files in SFTP server."""
        return super().__call__(*args, **kwargs)

    def run(
        self,
        path: str = None,
        recursive: bool = False,
        matching_path: str = None,
    ):
        logger = prefect.context.get("logger")

        if not self.credentials:
            if not self.sftp_credentials_secret:
                try:
                    self.sftp_credentials_secret = PrefectSecret("QWASI-SFTP").run()
                    logger.info("Loaded credentials from Key Vault.")
                except ValueError:
                    pass

            if self.sftp_credentials_secret:
                credentials_str = AzureKeyVaultSecret(
                    self.sftp_credentials_secret, vault_name=self.vault_name
                ).run()
                credentials_sftp = json.loads(credentials_str)
                logger.info("Loaded credentials from Key Vault.")
            else:
                credentials_sftp = local_config.get("QWASI-SFTP")
                logger.info("Loaded credentials from local source.")
        else:
            credentials_sftp = self.credentials

        sftp = SftpConnector(credentials_sftp=credentials_sftp)
        sftp.get_conn()
        logger.info("Connected to SFTP server.")
        if recursive is False:
            files_list = sftp.list_directory(path=path)

        else:
            files_list = sftp.recursive_listdir(path=path)
            files_list = sftp.process_defaultdict(defaultdict=files_list)
        if matching_path is not None:
            files_list = [f for f in files_list if matching_path in f]

        logger.info("Succefully loaded file list from SFTP server.")

        return files_list



from unittest import mock

import pandas as pd

from viadot.sources.mysql import MySQL

d = {"country": [1, 2], "sales": [3, 4]}
df = pd.DataFrame(data=d)

query = """SELECT * FROM `example-views`.`sales`"""


def test_create_mysql_instance():
    s = MySQL(credentials={"usr": 1, "pswd": 2})
    assert s


def test_connection_mysql():
    with mock.patch("viadot.sources.mysql.MySQL.to_df") as mock_method:
        mock_method.return_value = df
        s = MySQL(credentials={"usr": 1, "pswd": 2})

        final_df = s.to_df(query=query)
        assert type(final_df) == pd.DataFrame


def test_connect_mysql_ssh():
    with mock.patch("viadot.sources.mysql.MySQL.connect_sql_ssh") as mock_method:
        mock_method.return_value = df
        s = MySQL(credentials={"usr": 1, "pswd": 2})

        final_df = s.connect_sql_ssh(query=query)
        assert type(final_df) == pd.DataFrame



import pytest

from viadot.tasks import AzureKeyVaultSecret, CreateAzureKeyVaultSecret


class TestReadAzureKeyVaultSecret:
    def test_initialization(self):
        task = AzureKeyVaultSecret("test")

    def test_initialization_passes_to_task_constructor(self):
        task = AzureKeyVaultSecret(name="test", tags=["Azure"])
        assert task.name == "test"
        assert task.tags == {"Azure"}

    def test_raises_if_secret_not_eventually_provided(self):
        task = AzureKeyVaultSecret()

        with pytest.raises(ValueError, match="secret"):
            task.run()


class TestCreateAzureKeyVaultSecret:
    def test_initialization(self):
        task = CreateAzureKeyVaultSecret("test")

    def test_initialization_passes_to_task_constructor(self):
        task = CreateAzureKeyVaultSecret(name="test", tags=["Azure"])
        assert task.name == "test"
        assert task.tags == {"Azure"}

    def test_raises_if_secret_not_eventually_provided(self):
        task = CreateAzureKeyVaultSecret()

        with pytest.raises(ValueError, match="secret"):
            task.run()



from .azure_blob_storage import AzureBlobStorage
from .azure_data_lake import AzureDataLake
from .azure_sql import AzureSQL
from .bigquery import BigQuery
from .cloud_for_customers import CloudForCustomers
from .genesys import Genesys
from .mediatool import Mediatool
from .outlook import Outlook
from .salesforce import Salesforce
from .sftp import SftpConnector
from .sharepoint import Sharepoint
from .supermetrics import Supermetrics

try:
    from .sap_rfc import SAPRFC, SAPRFCV2
except ImportError:
    pass

try:
    from .sap_bw import SAPBW
except ImportError:
    pass

from .business_core import BusinessCore
from .customer_gauge import CustomerGauge
from .duckdb import DuckDB
from .epicor import Epicor
from .eurostat import Eurostat
from .hubspot import Hubspot
from .mindful import Mindful
from .sql_server import SQLServer
from .sqlite import SQLite

# APIS
from .uk_carbon_intensity import UKCarbonIntensity
from .vid_club import VidClub
